#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The main entrypoint for Poseidon

Created on 3 December 2018
@author: Charlie Lewis
"""
import logging
import sys
import threading
import time
from functools import partial

import schedule
from poseidon_core.controllers.faucet.config import FaucetRemoteConfGetSetter
from poseidon_core.controllers.sdnconnect import SDNConnect
from poseidon_core.controllers.sdnevents import SDNEvents
from poseidon_core.helpers.config import Config
from poseidon_core.helpers.log import Logger
from poseidon_core.helpers.prometheus import Prometheus
from poseidon_core.operations.monitor import Monitor


def start_prometheus(logger):
    prom = Prometheus()
    try:
        prom.initialize_metrics()
    except Exception as e:  # pragma: no cover
        logger.debug(
            f'Prometheus metrics are already initialized: {e}')
    Prometheus.start()
    return prom


def schedule_thread_worker(logger, scheduler=schedule):
    ''' schedule thread, takes care of running processes in the future '''
    logger.debug('Starting thread_worker')
    while True:
        sys.stdout.flush()
        scheduler.run_pending()
        time.sleep(1)


def main():  # pragma: no cover
    logging.getLogger('pika').setLevel(logging.CRITICAL)
    Logger()
    logger = logging.getLogger('main')
    config = Config().get_config()
    prom = start_prometheus(logger)

    # TODO option that doesn't require an sdn connection?
    sdnc = SDNConnect(config=config, logger=logger, prom=prom,
                      faucetconfgetsetter_cl=FaucetRemoteConfGetSetter)

    sdne = SDNEvents(logger, prom, sdnc)
    sdne.start_message_queues()

    # TODO this should be the default operation, but can be overridden with config to do other operations instead or additionally
    monitor = Monitor(logger, config, schedule, sdne.job_queue, sdnc, prom)

    # schedule all threads
    schedule_thread = threading.Thread(
        target=partial(
            schedule_thread_worker,
            logger, scheduler=schedule),
        name='st_worker')
    schedule_thread.start()

    try:
        # TODO each operation should have its own thread running its own "process" and this is just a main infinite loop
        sdne.process(monitor)
    except Exception as e:
        logger.exception(e)
        logger.error('restarting because of exception')
        sys.exit(1)



# -*- coding: utf-8 -*-
"""
Created on 9 December 2018
@author: Charlie Lewis
"""
from poseidon_core.helpers.collector import Collector
from poseidon_core.operations.volos.acls import VolosAcl


class Actions:

    def __init__(self, endpoint, sdnc):
        self.endpoint = endpoint
        self.sdnc = sdnc

    def mirror_endpoint(self):
        '''
        tell network_tap to start a collector and the controller to begin
        mirroring traffic
        '''
        status = False
        if self.sdnc:
            endpoint_data = self.endpoint.endpoint_data
            if self.sdnc.mirror_mac(endpoint_data['mac'], endpoint_data['segment'], endpoint_data['port']):
                collector = Collector(self.endpoint, endpoint_data['segment'])
                if collector.nic:
                    status = collector.start_collector()
        else:
            status = True
        return status

    def unmirror_endpoint(self):
        ''' tell the controller to unmirror traffic '''
        status = False
        if self.sdnc:
            endpoint_data = self.endpoint.endpoint_data
            if self.sdnc.unmirror_mac(endpoint_data['mac'], endpoint_data['segment'], endpoint_data['port']):
                collector = Collector(self.endpoint, endpoint_data['segment'])
                if collector.nic:
                    status = collector.stop_collector()
        else:
            status = True
        return status

    def coprocess_endpoint(self):
        '''
        Build up and apply acls for coprocessing
        '''
        status = False
        if self.sdnc:
            endpoint_data = self.endpoint.endpoint_data
            if self.sdnc.volos and self.sdnc.volos.enabled:
                acl = VolosAcl(
                    self.endpoint, acl_dir=self.sdnc.volos.acl_dir,
                    copro_vlans=[self.sdnc.volos.copro_vlan], copro_port=self.sdnc.volos.copro_port)
                endpoints = [self.endpoint]
                force_apply_rules = [acl.acl_key]
                coprocess_rules_files = [acl.acl_file]
                port_list = self.sdnc.volos.get_port_list(
                    endpoint_data['mac'], ipv4=endpoint_data.get('ipv4', None), ipv6=endpoint_data.get('ipv6', None))
                if acl.ensure_acls_dir() and acl.write_acl_file(port_list):
                    status = self.sdnc.update_acls(
                        rules_file=None, endpoints=endpoints,
                        force_apply_rules=force_apply_rules, coprocess_rules_files=coprocess_rules_files)
            else:
                status = True
        return status

    def uncoprocess_endpoint(self):
        ''' tell the controller to remove coprocessing acls'''
        status = False
        if self.sdnc:
            if self.sdnc.volos and self.sdnc.volos.enabled:
                acl = VolosAcl(
                    self.endpoint, acl_dir=self.sdnc.volos.acl_dir,
                    copro_vlans=[self.sdnc.volos.copro_vlan], copro_port=self.sdnc.volos.copro_port)
                endpoints = [self.endpoint]
                force_remove_rules = [acl.acl_key]
                if self.sdnc.update_acls(
                        rules_file=None, endpoints=endpoints,
                        force_remove_rules=force_remove_rules, coprocess_rules_files=None):
                    status = acl.delete_acl_file()
        else:
            status = True
        return status

    def update_acls(self, rules_file=None, endpoints=None, force_apply_rules=None, force_remove_rules=None):
        ''' tell the controller what ACLs to dynamically change '''
        status = False
        if self.sdnc:
            status = self.sdnc.update_acls(
                rules_file=rules_file, endpoints=endpoints,
                force_apply_rules=force_apply_rules, force_remove_rules=force_remove_rules)
        return status



# -*- coding: utf-8 -*-
"""
Created on 05 February 2020
@author: Ryan Ashley
"""
import logging
import os

import yaml
from poseidon_core.constants import PROTOCOL_MAP


class Volos(object):
    def __init__(self, controller):
        self.logger = logging.getLogger('volos')
        self.enabled = controller['enable_volos']
        self.coprocessor_nic = controller['coprocessor_nic']
        self.coprocessor_port = controller['coprocessor_port']
        self.coprocessor_vlans = controller['coprocessor_vlans']
        self.coprocessing_frequency = controller['coprocessing_frequency']
        self.ignore_copro_ports = controller['ignore_copro_ports']
        self.acl_dir = controller['acl_dir']
        volos_cfg_file = controller['volos_cfg_file']
        self.container_config = self.parse_volos_cfg(volos_cfg_file)

    def parse_volos_cfg(self, volos_cfg_file):
        cfg = None
        container_cfg = None
        if os.path.exists(volos_cfg_file):
            try:
                with open(volos_cfg_file, 'r') as f:
                    cfg = yaml.safe_load(f)
            except Exception as e:  # pragma: no cover
                self.logger.warning(
                    'Volos configuration could not be loaded, disabling Volos')
                self.logger.error(
                    'Failed to load Volos config with error: {0}'.format(str(e)))
                self.enabled = False
            container_cfg = []
            if cfg:
                for repo in cfg:
                    item = {}
                    for name in cfg[repo]:
                        item['repo'] = repo
                        item['name'] = name
                        item['branch'] = cfg[repo][name]['branch']
                        item['build'] = cfg[repo][name]['build']
                        item['start'] = cfg[repo][name]['start']
                        item['ports'] = []
                        for port in cfg[repo][name]['ports']:
                            cfg_p = {}
                            cfg_p['proto'] = port['port']['protocol']
                            cfg_p['proto_id'] = PROTOCOL_MAP[port['port']
                                                             ['protocol']]
                            mapping = port['port']['mapping']
                            cfg_p['host'] = mapping[:mapping.index(':')]
                            cfg_p['dest'] = mapping[mapping.index(':'):]
                            item['ports'].append(cfg_p)

                    container_cfg.append(item)

            else:
                self.enabled = False
        else:
            self.logger.debug(
                'Volos configuration could not be found. disabling volos')
            self.enabled = False

        return container_cfg

    '''
    build structure of the form
    {
    'mac1': {
        'ip': {
         'v4': "ipv4_1",
         'v6': "ipv6_1",
        },
        'ports': [
            {
                'proto': 'tcp',
                'proto_id': 6,
                'port': 25,
            },
            {
                'proto': 'tcp',
                'proto_id': 6,
                'port': 26,
            },
            {
                'proto': 'udp',
                'proto_id': 17,
                'port': 27,
            }
        ]
    },
    '''

    def get_port_list(self, mac, ipv4=None, ipv6=None):
        port_list = {}
        port_list[mac] = {
            'ip': {
                'v4': ipv4,
                'v6': ipv6
            },
            'ports': []
        }
        for i in self.container_config:
            for port in i['ports']:
                p = {}
                p['proto'] = port['proto']
                p['proto_id'] = port['proto_id']
                p['port'] = port['dest']

                port_list[mac]['ports'].append(p)

        return port_list



from poseidon_core.helpers.config import Config


class Mirror:

    def __init__(self, logger):
        config = Config().get_config()
        self.logger = logger
        self.mirror_ports = config['MIRROR_PORTS']
        self.proxy_mirror_ports = config['controller_proxy_mirror_ports']
        self.tunnel_vlan = config['tunnel_vlan']
        self.tunnel_name = config['tunnel_name']
        self.ignore_vlans = config['ignore_vlans']
        self.ignore_ports = config['ignore_ports']
        self.trunk_ports = config['truck_ports']

    def mirror_port(self, switch, port):
        self.logger.warning(f'Unable to mirror {switch}:{port}')
        return False

    def unmirror_port(self, switch, port):
        self.logger.warning(f'Unable to unmirror {switch}:{port}')
        return False

    def mirror_mac(self, switch, port, mac):
        self.logger.warning(f'Unable to mirror {switch}:{port}:{mac}')
        return False

    def unmirror_mac(self, switch, port, mac):
        self.logger.warning(f'Unable to unmirror {switch}:{port}:{mac}')
        return False

    def mirror_endpoint(self, endpoint):
        self.logger.warning(f'Unable to mirror {endpoint.name}')
        return False

    def unmirror_endpoint(self, endpoint):
        self.logger.warning(f'Unable to unmirror {endpoint.name}')
        return False

    def clear_mirrors(self):
        self.logger.warning('Unable to clear all mirrors')
        return False



# -*- coding: utf-8 -*-
"""
Created on 27 December 2018
@author: Charlie Lewis
"""
import functools
import logging


def exception(function):
    """
    A decorator that wraps the passed in function and logs
    exceptions should one occur
    """
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        logger = logging.getLogger('exception')
        try:
            return function(*args, **kwargs)
        except Exception as e:
            # log the exception
            logger.exception('Exception in {0}: {1}'.format(
                function.__name__, str(e)))
            return False
    return wrapper



# -*- coding: utf-8 -*-
"""
Created on 4 March 2020
@author: Charlie Lewis
"""
import logging
import os


class ACL:

    def __init__(self, faucetconfgetsetter):
        self.logger = logging.getLogger('acl')
        self.frpc = faucetconfgetsetter

    def _config_file_paths(self, file_paths):
        return [self.frpc.config_file_path(f) for f in file_paths]

    def include_acl_files(self, rules_doc, rules_file, coprocess_rules_files, obj_doc):
        files = self._config_file_paths(rules_doc['include'])
        rules_path = rules_file.rsplit('/', 1)[0]
        conf_files = obj_doc.get('include', [])

        acls_docs = {}
        for f in files:
            if f.startswith('/'):
                acls_doc = self.frpc.read_faucet_conf(f)
            else:
                acls_doc = self.frpc.read_faucet_conf(
                    os.path.join(rules_path, f))
            if isinstance(acls_doc, bool):
                self.logger.warning(
                    'Include file {0} was not found, ACLs may not be working as expected'.format(f))
                continue
            acls_docs[f] = acls_doc

        if conf_files:
            acls_filenames = []
            if coprocess_rules_files:
                acls_filenames += self._config_file_paths(
                    coprocess_rules_files)
            for f in files:
                if '/' in f:
                    acls_filenames.append(f.rsplit('/', 1)[1])
                else:
                    acls_filenames.append(f)
            for conf_file in conf_files:
                if conf_file.startswith('poseidon') and conf_file not in acls_filenames:
                    obj_doc['include'].remove(conf_file)
                    self.logger.info(
                        'Removing {0} from config'.format(conf_file))
        else:
            obj_doc['include'] = []

        for f, acls_doc in acls_docs.items():
            if '/' in f:
                _, acls_filename = f.rsplit('/', 1)
            else:
                acls_filename = f
            poseidon_acls_filename = 'poseidon_' + acls_filename
            if poseidon_acls_filename not in conf_files:
                obj_doc['include'].append(poseidon_acls_filename)
                self.frpc.write_faucet_conf(os.path.join(
                    rules_path, poseidon_acls_filename), acls_doc)
                self.logger.info('Adding {0} to config'.format(acls_filename))

        # get defined ACL names from included files
        acl_names = []
        for acls_doc in acls_docs.values():
            acl_names.extend(list(acls_doc.get('acls', [])))
        return obj_doc, acl_names

    def match_rules(self, rule, rules, obj_doc, endpoint, switch, port, all_rule_acls, force_apply_rules):
        matches = 0
        for r in rules[rule]:
            if 'rule' in r and 'device_key' in r['rule']:
                rule_data = r['rule']
                if rule_data['device_key'] == 'os':
                    match = False
                    for addresses in ('ipv4_addresses', 'ipv6_addresses'):
                        if 'addresses' in endpoint.metadata:
                            for ip, ip_metadata in endpoint.metadata['addresses']:
                                if 'os' in ip_metadata and ip_metadata['os'] == rule_data['value']:
                                    self.logger.info('{0} os match: {1} {2}, rule: {3}'.format(
                                        addresses, ip, rule_data['value'], rule))
                                    match = True
                    if match:
                        matches += 1
                elif rule_data['device_key'] == 'role':
                    match = False
                    if 'mac_addresses' in endpoint.metadata:
                        for mac, mac_metadata in endpoint.metadata['mac_addresses'].items():
                            most_recent = 0
                            for record in mac_metadata:
                                if float(record) > most_recent:
                                    most_recent = float(record)
                            most_recent = str(most_recent)
                            if most_recent != '0' and most_recent in mac_metadata and 'labels' in mac_metadata[most_recent] and 'confidences' in mac_metadata[most_recent]:
                                # check top three
                                for i in range(3):
                                    if mac_metadata[most_recent]['labels'][i] == rule_data['value']:
                                        if 'min_confidence' in rule_data['value']:
                                            if float(mac_metadata[most_recent]['confidences'][i])*100 >= rule_data['min_confidence']:
                                                self.logger.info('Confidence match: {0} {1}, rule: {2}'.format(mac, float(
                                                    mac_metadata[most_recent]['confidences'][i])*100, rule))
                                                match = True
                                        else:
                                            self.logger.info('Role match: {0} {1}, rule: {2}'.format(
                                                mac, rule_data['value'], rule))
                                            match = True
                    if match:
                        matches += 1
        if matches == len(rules[rule]) or (force_apply_rules and rule in force_apply_rules):
            rule_acls = []
            for r in rules[rule]:
                rule_acls += r['rule']['acls']
                all_rule_acls += r['rule']['acls']
            rule_acls = list(set(rule_acls))
            interfaces_conf = obj_doc['dps'][switch]['interfaces']
            if rule_acls:
                if port not in interfaces_conf:
                    interfaces_conf[port] = {}
            port_conf = interfaces_conf[port]
            if 'acls_in' not in port_conf:
                self.logger.info('All rules met for: {0} on switch: {1} and port: {2}; applying ACLs: {3}'.format(
                    endpoint.endpoint_data['mac'], switch, port, rule_acls))
                port_conf['acls_in'] = rule_acls
            else:
                # add new ACLs
                orig_rule_acls = rule_acls
                rule_acls += port_conf['acls_in']
                rule_acls = list(set(rule_acls))
                if port_conf['acls_in'] != rule_acls:
                    port_conf['acls_in'] = rule_acls
                    self.logger.info('All rules met for: {0} on switch: {1} and port: {2}; applying ACLs: {3}'.format(
                        endpoint.endpoint_data['mac'], switch, port, orig_rule_acls))
        return obj_doc, all_rule_acls

    def apply_acls(self, rules_file, endpoints, force_apply_rules,
                   force_remove_rules, coprocess_rules_files, obj_doc,
                   rules_doc):
        if not endpoints:
            return obj_doc

        # get acls file and add to faucet.yaml if not already there
        if 'include' not in rules_doc:
            self.logger.info(
                'No included ACLs files in the rules file, using ACLs that Faucet already knows about')
        else:
            obj_doc, acl_names = self.include_acl_files(rules_doc, rules_file,
                                                        coprocess_rules_files,
                                                        obj_doc)

        if 'rules' in rules_doc:
            acls = []
            rules = rules_doc['rules']

            for rule in rules:
                for r in rules[rule]:
                    acls += r['rule']['acls']
            acls = list(set(acls))

            # check that acls in rules exist in the included acls file
            if 'include' in rules_doc:
                for acl in acls:
                    if acl not in acl_names:
                        self.logger.info(
                            'Using named ACL: {0}, but it was not found in included ACL files, assuming ACL name exists in Faucet config'.format(acl))

            for endpoint in endpoints:
                port = int(endpoint.endpoint_data['port'])
                switch = endpoint.endpoint_data['segment']
                switch_conf = obj_doc['dps'].get(switch, None)
                if not switch_conf:
                    continue
                port_conf = switch_conf['interfaces'].get(port, None)
                if not port_conf:
                    continue
                existing_acls = port_conf.get('acls_in', None)
                if not existing_acls:
                    continue
                all_rule_acls = []
                for rule in rules:
                    obj_doc, all_rule_acls = self.match_rules(
                        rule, rules, obj_doc, endpoint, switch, port, all_rule_acls, force_apply_rules)
                # remove ACLs that were previously applied
                all_rule_acls = list(set(all_rule_acls))
                for acl in existing_acls:
                    if acl in acls and (acl not in all_rule_acls or acl in force_remove_rules):
                        port_conf['acls_in'].remove(acl)
                        self.logger.info('Removing no longer needed ACL: {0} for: {1} on switch: {2} and port: {3}'.format(
                            acl, endpoint.endpoint_data['mac'], switch, port))

        # TODO acl by port - potentially later update rules in acls to be mac/ip specific
        # TODO ignore trunk ports/stacking ports?

        return obj_doc



# -*- coding: utf-8 -*-
'''
Created on 17 November 2017
@author: Charlie Lewis
'''
import ipaddress
import logging
from collections import defaultdict

from poseidon_core.controllers.faucet.config import FaucetRemoteConfGetSetter
from poseidon_core.helpers.config import parse_rules
from poseidon_core.operations.primitives.acl import ACL
from poseidon_core.operations.primitives.coprocess import Coprocess
from poseidon_core.operations.volos.acls import Acl
from poseidon_core.operations.volos.volos import Volos


class FaucetProxy:

    def __init__(self,
                 config,
                 *args,
                 **kwargs):
        '''Initializes Faucet object.'''
        self.mirror_ports = kwargs.get('mirror_ports', config['MIRROR_PORTS'])
        self.proxy_mirror_ports = kwargs.get(
            'proxy_mirror_ports', config['controller_proxy_mirror_ports'])
        self.tunnel_vlan = kwargs.get('tunnel_vlan', config['tunnel_vlan'])
        self.tunnel_name = kwargs.get('tunnel_name', config['tunnel_name'])
        self.learn_pub_adds = kwargs.get(
            'learn_pub_adds', config['LEARN_PUBLIC_ADDRESSES'])
        self.reinvestigation_frequency = kwargs.get(
            'reinvestigation_frequency', config['reinvestigation_frequency'])
        self.max_concurrent_reinvestigations = kwargs.get(
            'max_concurrent_reinvestigations', config['max_concurrent_reinvestigations'])
        self.trunk_ports = kwargs.get('trunk_ports', config['trunk_ports'])
        self.ignore_vlans = kwargs.get('ignore_vlans', config['ignore_vlans'])
        self.ignore_ports = kwargs.get('ignore_ports', config['ignore_ports'])
        self.mirror_counts = kwargs.get('mirror_counts', defaultdict(int))
        self.frpc = None
        faucetconfgetsetter_cl = kwargs.get(
            'faucetconfgetsetter_cl', FaucetRemoteConfGetSetter)
        self._get_frpc(config, faucetconfgetsetter_cl=faucetconfgetsetter_cl)
        self._set_default_switch_conf()
        self.logger = logging.getLogger('faucet')
        self.mac_table = {}

        # parse volos config
        self.volos = Volos(config)
        self.coprocessor = Coprocess(config)

    def _get_acls(self):
        faucet_conf = self.frpc.read_faucet_conf(config_file=None)
        acls = Acl(faucetconfgetsetter=self.frpc)
        acls.read(config_yaml=faucet_conf)
        return acls

    def _set_default_switch_conf(self):
        # TODO: make smarter with more complex configs (backup original values, etc)
        if self.mirror_ports:
            root_stack_switch = self.frpc.get_stack_root_switch()
            root_mirror_port = None
            if root_stack_switch:
                acls = self._get_acls()
                root_mirror_port = self.mirror_ports.get(
                    root_stack_switch, None)
                if self.tunnel_name in acls.acls:
                    del acls.acls[self.tunnel_name]
                # Safety rule to prevent looped packets being output to the loop.
                acls.add_rule(
                    self.tunnel_name,
                    {'rule': {'vlan_vid': self.tunnel_vlan, 'actions': {'allow': 0}}})
                # Tunnel back to root.
                acls.add_rule(
                    self.tunnel_name,
                    {'rule': {'actions': {'allow': 0, 'output': {'tunnel': {
                        'type': 'vlan', 'tunnel_id': self.tunnel_vlan,
                        'dp': root_stack_switch, 'port': root_mirror_port}}}}})
                # Merge ACL updates, if any back in
                self.frpc.set_acls(acls.acls)

            for switch, mirror_port in self.mirror_ports.items():
                if not self.frpc.get_switch_conf(switch):
                    continue
                if self.reinvestigation_frequency:
                    switch_conf = {
                        'timeout': (self.reinvestigation_frequency * 2) + 1,
                        'arp_neighbor_timeout': self.reinvestigation_frequency}
                    self.frpc.update_switch_conf(switch, switch_conf)
                # If stacking was detected, provision tunnel config on non root switches.
                # Poseidon's mirror NIC must be connected to the root switch's mirror port.
                # Non mirror switches must have a loopback plug installed in their mirror port.
                if root_stack_switch and root_mirror_port:
                    mirror_port_conf = self.frpc.get_port_conf(
                        switch, mirror_port)
                    if not mirror_port_conf:
                        continue
                    for existing_key in set(mirror_port_conf.keys()):
                        if existing_key not in ('mirror',):
                            del mirror_port_conf[existing_key]
                    if root_stack_switch == switch:
                        mirror_port_conf.update({
                            'description': 'Poseidon local mirror',
                            'output_only': True,
                        })
                    else:
                        mirror_port_conf.update({
                            'description': 'Poseidon remote mirror (loopback plug)',
                            'acls_in': [self.tunnel_name],
                            'coprocessor': {'strategy': 'vlan_vid'},
                        })
                    self.frpc.set_port_conf(
                        switch, mirror_port, mirror_port_conf)

    def _get_frpc(self, config, faucetconfgetsetter_cl=FaucetRemoteConfGetSetter):
        faucetconfrpc_address = config['faucetconfrpc_address']
        faucetconfrpc_client = config['faucetconfrpc_client']
        if faucetconfrpc_address:
            server = faucetconfrpc_address.split(':')[0]
        else:
            server = ''
        self.frpc = faucetconfgetsetter_cl(
            client_key='/certs/%s.key' % faucetconfrpc_client,
            client_cert='/certs/%s.crt' % faucetconfrpc_client,
            ca_cert='/certs/%s-ca.crt' % server,
            server_addr=faucetconfrpc_address)

    @staticmethod
    def format_endpoints(data):
        '''
        return only the information needed for the application
        '''
        ret_list = list()
        for d in data:
            md = d[0]
            d.reverse()
            for i, _ in enumerate(d):
                ipv4_set = False
                ipv6_set = False
                if 'ip-address' in d[i]:
                    if ':' in d[i]['ip-address']:
                        md['ipv6'] = d[i]['ip-address']
                        ipv6_set = True
                    else:
                        md['ipv4'] = d[i]['ip-address']
                        ipv4_set = True
                if 'ipv4' in md:
                    ipv4_set = True
                if 'ipv6' in md:
                    ipv6_set = True
            if not ipv4_set:
                md['ipv4'] = 0
            if not ipv6_set:
                md['ipv6'] = 0
            if 'ip-state' in md:
                del md['ip-state']
            if 'ip-address' in md:
                del md['ip-address']

            md.update({
                'controller_type': 'faucet',
                'controller': '',
                'name': None})
            ret_list.append(md)
        return ret_list

    def ignore_event(self, message):
        for message_type in ('L2_LEARN',):
            message_body = message.get(message_type, None)
            if message_body:
                switch = str(message['dp_name'])
                port_no = message_body.get('port_no', None)
                vlan = message_body.get('vid', None)
                # When stacking is in use, we only want to learn on switch, that a host is local to.
                if 'stack_descr' in message_body:
                    self.logger.debug(
                        'Ignoring event because learning from a stack port')
                    return True
                if self.ignore_vlans:
                    if vlan in self.ignore_vlans:
                        self.logger.debug(
                            'Ignoring event because VLAN %s ignored' % vlan)
                        return True
                if self.ignore_ports:
                    for ignore_switch, ignore_port_no in self.ignore_ports.items():
                        if ignore_switch == switch and ignore_port_no == port_no:
                            self.logger.debug(
                                'Ignoring event because switch %s port %s is ignored' % (
                                    switch, port_no))
                            return True
                if self.proxy_mirror_ports:
                    for s in self.proxy_mirror_ports:
                        if (switch == self.proxy_mirror_ports[s][0] and
                                port_no == self.proxy_mirror_ports[s][1]):
                            self.logger.debug(
                                'Ignoring event because switch %s port %s is being a proxy' % (
                                    switch, port_no))
                            return True
                # Not on any ignore list, don't ignore.
                return False
        # Not a message we are interested in, ignore it.
        return True

    def event(self, message):
        dp_name = str(message['dp_name'])

        if 'L2_LEARN' in message:
            self.logger.debug(
                'Got faucet message for l2_learn: {0}'.format(message))
            message = message['L2_LEARN']
            eth_src = message['eth_src']
            vlan_str = 'VLAN%s' % message['vid']
            data = {
                'ip-address': message['l3_src_ip'],
                'ip-state': 'L2 learned',
                'mac': eth_src,
                'segment': dp_name,
                'port': str(message['port_no']),
                'vlan': vlan_str,
                'tenant': vlan_str,
                'active': 1}

            if eth_src in self.mac_table:
                if data in self.mac_table[eth_src]:
                    self.mac_table[eth_src].remove(data)
                self.mac_table[eth_src].insert(0, data)
            else:
                self.mac_table[eth_src] = [data]

    def get_endpoints(self, messages=None):
        retval = []

        if messages:
            self.logger.debug('faucet messages: {0}'.format(messages))
            for message in messages:
                if not self.ignore_event(message):
                    self.event(message)
        for mac in self.mac_table:
            if self.learn_pub_adds:
                retval.append(self.mac_table[mac])
            else:
                # only allow private addresses
                first_entry = self.mac_table[mac][0]
                if 'ip-address' in first_entry and (
                        first_entry['ip-address'] == 'None' or
                        first_entry['ip-address'] is None or
                        not ipaddress.ip_address(first_entry['ip-address']).is_global):
                    retval.append(self.mac_table[mac])
        return retval

    def update_acls(self, rules_file=None, endpoints=None, force_apply_rules=None, force_remove_rules=None):
        self.logger.debug('Updating ACLs')
        rules_doc = parse_rules(rules_file)
        if not rules_doc:
            self.logger.error(
                'Unable to read or parse rules file, not applying ACLs')
            return False
        self.frpc.read_faucet_conf(config_file=None)
        # TODO coprocess_rules_files is set to None, which was previous default but removes functionality
        self.frpc.faucet_conf = ACL(self.frpc).apply_acls(
            rules_file, endpoints,
            force_apply_rules, force_remove_rules,
            None, self.frpc.faucet_conf, rules_doc)
        self.frpc.write_faucet_conf(config_file=None)
        return True

    def _mac_switch_port(self, my_mac):
        try:
            entry = self.mac_table[my_mac][0]
            return (entry['segment'], int(entry['port']))
        except (KeyError, IndexError):
            return (None, None)

    def proxy_mirror_port(self, switch, port):
        self.logger.debug('Checking for proxy ports')
        if self.proxy_mirror_ports and self.frpc.get_switch_conf(switch):
            proxy_ports = self.proxy_mirror_ports.get(switch, None)
            if proxy_ports:
                self.logger.debug(f'Found proxy port: {switch} port {port}')
                switch, port = proxy_ports
        return switch, port

    def mirror_switch_port(self, switch):
        if self.mirror_ports:
            switch_mirror_port = self.mirror_ports.get(switch, None)
            if switch_mirror_port:
                if self.frpc.get_port_conf(switch, switch_mirror_port):
                    return switch_mirror_port
        self.logger.warning('No mirror port for switch %s' % switch)
        return None

    def clear_mirrors(self):
        dps = self.frpc.get_dps()
        if dps:
            for switch in dps:
                mirror_port = self.mirror_switch_port(switch)
                if mirror_port:
                    self.frpc.clear_mirror_port(switch, mirror_port)

    def mirror_mac(self, my_mac, my_switch, my_port):
        self.logger.debug('Mirroring mac %s', my_mac)
        switch, port = self._mac_switch_port(my_mac)
        if switch and port:
            switch, port = self.proxy_mirror_port(switch, port)
            mirror_port = self.mirror_switch_port(switch)
            if mirror_port:
                mirror_key = (switch, port)
                self.logger.info(f'Request mirror of {mirror_key}')
                self.frpc.mirror_port(switch, mirror_port, port)
                self.mirror_counts[mirror_key] += 1
                count = self.mirror_counts[mirror_key]
                self.logger.info(f'Mirroring {count} MACs on {mirror_key}')
            else:
                self.logger.error(
                    f'Unable to configure mirror on {switch}:{port} due to warnings')
                return False
        return True

    def unmirror_mac(self, my_mac, my_switch, my_port):
        self.logger.debug('Unmirroring mac %s', my_mac)
        switch, port = self._mac_switch_port(my_mac)
        if port and switch:
            trunk = False
            for sw in self.trunk_ports:
                if sw == switch and self.trunk_ports[sw].split(',')[1] == str(port):
                    trunk = True
            if not trunk:
                switch, port = self.proxy_mirror_port(switch, port)
                mirror_port = self.mirror_switch_port(switch)
                if mirror_port:
                    mirror_key = (switch, port)
                    self.logger.info(f'Request unmirror of {mirror_key}')
                    if self.mirror_counts[mirror_key]:
                        self.mirror_counts[mirror_key] -= 1
                        if not self.mirror_counts[mirror_key]:
                            self.logger.info(
                                f'Removing last remaining mirror on {mirror_key}')
                            self.frpc.unmirror_port(switch, mirror_port, port)
                        count = self.mirror_counts[mirror_key]
                        self.logger.info(
                            f'Mirroring {count} MACs on {mirror_key}')
                else:
                    self.logger.error(
                        f'Unable to configure unmirror on {switch}:{port} due to warnings')
                    return False
            else:
                self.logger.debug('Not unmirroring a trunk port')
                return False
        return True

    def coprocess_mac(self, my_mac):
        self.logger.debug('coprocess mac: {0}'.format(my_mac))
        switch, port = self._mac_switch_port(my_mac)
        if switch and port:
            # TODO this is an unknown action currently...
            #self.config('coprocess', int(port), switch)
            pass
        return True

    def uncoprocess_mac(self, my_mac):
        self.logger.debug('uncoprocess mac: {0}'.format(my_mac))
        switch, port = self._mac_switch_port(my_mac)
        if switch and port:
            # TODO this is an unknown action currently...
            #self.config('uncoprocess', int(port), switch)
            pass
        return True



# -*- coding: utf-8 -*-
"""
Created on 18 September 2017
@author: Jeff Wang, Charlie Lewis
"""
import logging.handlers
import os
import socket

from poseidon_core.helpers.config import Config


class Logger:
    """
    Base logger class that handles logging. Outputs to both console, a poseidon
    specific log file and a user specified syslog. To log, create a logger:
    logger1 = logging.getLogger('mpapp.area1')
    """
    host = os.getenv('SYSLOG_HOST', 'NOT_CONFIGURED')
    port = int(os.getenv('SYSLOG_PORT', 514))

    level_int = {'CRITICAL': 50, 'ERROR': 40, 'WARNING': 30,
                 'INFO': 20, 'DEBUG': 10}

    config = Config().get_config()

    # setup existing loggers
    logging.getLogger('schedule').setLevel(logging.ERROR)

    use_file_logger = True
    # ensure log file exists
    try:
        if not os.path.exists('/var/log/poseidon'):
            os.makedirs('/var/log/poseidon')
        if not os.path.exists('/var/log/poseidon/poseidon.log'):
            with open('/var/log/poseidon/poseidon.log', 'w'):
                pass
        # set up logging to file
        level_str = config.get('logger_level', None)
        level = 0
        if isinstance(level_str, str):
            level = level_int.get(level_str.upper(), 0)
        logging.basicConfig(level=level,
                            format='%(asctime)s [%(levelname)s] %(name)s - %(message)s',
                            filename='/var/log/poseidon/poseidon.log',
                            filemode='a')
    except Exception as e:  # pragma: no cover
        use_file_logger = False

    # define a Handler which writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    # set a format which is simpler for console use
    formatter = logging.Formatter('[%(levelname)s] %(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    logging.getLogger('').addHandler(console)

    # don't try to connect to a syslog address if one was not supplied
    if host != 'NOT_CONFIGURED':  # pragma: no cover
        # if a syslog address was supplied, log to it
        syslog = logging.handlers.SysLogHandler(
            address=(host, port), socktype=socket.SOCK_STREAM)
        f_format = '%(asctime)s [%(levelname)s] %(name)s - %(message)s'
        f_formatter = logging.Formatter(f_format)
        syslog.setFormatter(f_formatter)
        logging.getLogger('').addHandler(syslog)



import logging
import os
import sys

import yaml
from faucetconfrpc.faucetconfrpc_client_lib import FaucetConfRpcClient


class FaucetRemoteConfGetSetter:

    DEFAULT_CONFIG_FILE = ''

    def __init__(self, client_key=None, client_cert=None,
                 ca_cert=None, server_addr=None):
        self.client = FaucetConfRpcClient(
            client_key=client_key, client_cert=client_cert,
            ca_cert=ca_cert, server_addr=server_addr)

    @staticmethod
    def config_file_path(config_file):
        if config_file:
            return os.path.basename(config_file)
        return config_file

    def read_faucet_conf(self, config_file):
        self.faucet_conf = self.client.get_config_file(
            config_filename=self.config_file_path(config_file))
        if self.faucet_conf is None:
            logging.error('Faucet config is empty, exiting.')
            sys.exit(1)
        return self.faucet_conf

    def write_faucet_conf(self, config_file=None, faucet_conf=None, merge=False):
        if not config_file:
            config_file = self.DEFAULT_CONFIG_FILE
        if faucet_conf is None:
            faucet_conf = self.faucet_conf
        return self.client.set_config_file(
            self.faucet_conf,
            config_filename=self.config_file_path(config_file),
            merge=merge)

    def get_dps(self):
        self.read_faucet_conf(config_file=None)
        return self.faucet_conf.get('dps', {})

    def set_acls(self, acls):
        self.read_faucet_conf(config_file=None)
        self.faucet_conf['acls'] = acls
        self.write_faucet_conf(config_file=None)

    def get_port_conf(self, dp, port):
        switch_conf = self.get_switch_conf(dp)
        if not switch_conf:
            return None
        return switch_conf['interfaces'].get(port, None)

    def get_switch_conf(self, dp):
        return self.get_dps().get(dp, None)

    def get_stack_root_switch(self):
        root_stack_switch = [
            switch for switch, switch_conf in self.get_dps().items()
            if switch_conf.get('stack', {}).get('priority', None)]
        if root_stack_switch:
            return root_stack_switch[0]
        return None

    def set_port_conf(self, dp, port, port_conf):
        return self.client.set_dp_interfaces(
            [(dp, {port: yaml.dump(port_conf)})])

    def update_switch_conf(self, dp, switch_conf):
        return self.write_faucet_conf(
            faucet_conf={'dps': {dp: switch_conf}}, merge=True)

    def mirror_port(self, dp, mirror_port, port):  # pragma: no cover
        self.client.add_port_mirror(dp, port, mirror_port)

    def unmirror_port(self, dp, mirror_port, port):  # pragma: no cover
        self.client.remove_port_mirror(dp, port, mirror_port)

    def clear_mirror_port(self, dp, mirror_port):  # pragma: no cover
        self.client.clear_port_mirror(dp, mirror_port)



# -*- coding: utf-8 -*-
"""
Created on 3 December 2018
@author: Charlie Lewis
"""
import hashlib
import json
import time

from poseidon_core.constants import NO_DATA
from transitions import Machine

MACHINE_IP_FIELDS = {
    'ipv4': ('ipv4_rdns', 'ipv4_subnet'),
    'ipv6': ('ipv6_rdns', 'ipv6_subnet')}
MACHINE_IP_PREFIXES = {
    'ipv4': 24, 'ipv6': 64}


def transit_wrap(trigger, source, dest, before=None, after=None):
    transit_dict = {'trigger': trigger, 'source': source, 'dest': dest}
    if before is not None:
        transit_dict['before'] = before
    if after is not None:
        transit_dict['after'] = after
    return transit_dict


def endpoint_transit_wrap(trigger, source, dest):
    return transit_wrap(trigger, source, dest, after='_update_state_time')


def endpoint_copro_transit_wrap(trigger, source, dest):
    return transit_wrap(trigger, source, dest, after='_update_copro_state_time')


class Endpoint:

    states = ['known', 'unknown', 'operating', 'queued']

    transitions = [
        endpoint_transit_wrap('operate', 'unknown', 'operating'),
        endpoint_transit_wrap('queue', 'unknown', 'queued'),
        endpoint_transit_wrap('operate', 'known', 'operating'),
        endpoint_transit_wrap('queue', 'known', 'queued'),
        endpoint_transit_wrap('operate', 'queued', 'operating'),
        endpoint_transit_wrap('known', 'known', 'known'),
        endpoint_transit_wrap('unknown', 'known', 'unknown'),
        endpoint_transit_wrap('known', 'unknown', 'known'),
        endpoint_transit_wrap('unknown', 'unknown', 'unknown'),
        endpoint_transit_wrap('known', 'operating', 'known'),
        endpoint_transit_wrap('unknown', 'operating', 'unknown'),
        endpoint_transit_wrap('known', 'queued', 'known'),
        endpoint_transit_wrap('unknown', 'queued', 'unknown'),
    ]

    copro_states = ['copro_unknown', 'copro_coprocessing',
                    'copro_nominal', 'copro_suspicious', 'copro_queued']

    copro_transitions = [
        endpoint_copro_transit_wrap(
            'copro_coprocess', 'copro_unknown', 'copro_coprocessing'),
        endpoint_copro_transit_wrap(
            'copro_queue', 'copro_unknown', 'copro_queued'),
        endpoint_copro_transit_wrap(
            'copro_coprocess', 'copro_queued', 'copro_coprocessing'),
        endpoint_copro_transit_wrap(
            'copro_nominal', 'copro_coprocessing', 'copro_nominal'),
        endpoint_copro_transit_wrap(
            'copro_nominal', 'copro_queued', 'copro_nominal'),
        endpoint_copro_transit_wrap(
            'copro_suspicious', 'copro_coprocessing', 'copro_suspicious'),
        endpoint_copro_transit_wrap(
            'copro_queue', 'copro_nominal', 'copro_queued'),
        endpoint_copro_transit_wrap(
            'copro_coprocess', 'copro_nominal', 'copro_coprocessing'),
        endpoint_copro_transit_wrap(
            'copro_queue', 'copro_suspicious', 'copro_queued'),
        endpoint_copro_transit_wrap(
            'copro_coprocess', 'copro_suspicious', 'copro_coprocessing'),
    ]

    def __init__(self, hashed_val):
        self.name = hashed_val.strip()
        self.ignore = False
        self.copro_ignore = False
        self.endpoint_data = None
        self.p_next_state = None
        self.p_prev_state = None
        self.p_next_copro_state = None
        self.p_prev_copro_state = None
        self.acl_data = []
        self.metadata = {}
        self.state = None
        self.copro_state = None
        self.state_time = 0
        self.copro_state_time = 0
        self.observed_time = 0

    def _update_state_time(self, *args, **kwargs):
        self.state_time = time.time()

    def _update_copro_state_time(self, *args, **kwargs):
        self.copro_state_time = time.time()

    def encode(self):
        endpoint_d = {
            'name': self.name,
            'state': self.state,
            'copro_state': self.copro_state,
            'ignore': self.ignore,
            'endpoint_data': self.endpoint_data,
            'p_next_state': self.p_next_state,
            'p_prev_state': self.p_prev_state,
            'acl_data': self.acl_data,
            'metadata': self.metadata,
            'observed_time': self.observed_time,
        }
        return str(json.dumps(endpoint_d))

    def mac_addresses(self):
        return self.metadata.get('mac_addresses', {})

    def get_roles_confidences_pcap_labels(self):
        top_role = NO_DATA
        second_role = NO_DATA
        third_role = NO_DATA
        top_conf = '0'
        second_conf = '0'
        third_conf = '0'
        pcap_labels = NO_DATA
        for metadata in self.mac_addresses().values():
            classification = metadata.get('classification', {})
            if 'labels' in classification:
                top_role, second_role, third_role = classification['labels'][:3]
            if 'confidences' in classification:
                top_conf, second_conf, third_conf = classification['confidences'][:3]
            metadata_pcap_labels = metadata.get('pcap_labels', None)
            if metadata_pcap_labels:
                pcap_labels = metadata_pcap_labels
        return (top_role, second_role, third_role), (top_conf, second_conf, third_conf), pcap_labels

    def get_ipv4_os(self):
        if 'ipv4_addresses' in self.metadata:
            ipv4 = self.endpoint_data['ipv4']
            for ip, ip_metadata in self.metadata['ipv4_addresses'].items():
                if ip == ipv4:
                    if 'short_os' in ip_metadata:
                        return ip_metadata['short_os']
        return NO_DATA

    def touch(self):
        self.observed_time = time.time()

    def observed_timeout(self, timeout):
        return time.time() - self.observed_time > timeout

    def state_age(self):
        return int(time.time()) - self.state_time

    def state_timeout(self, timeout):
        return self.state_age() > timeout

    def copro_state_age(self):
        return int(time.time()) - self.copro_state_time

    def copro_state_timeout(self, timeout):
        return self.copro_state_age() > timeout

    def queue_next(self, next_state):
        self.p_next_state = next_state
        self.queue()  # pytype: disable=attribute-error

    def machine_trigger(self, state):
        # pytype: disable=attribute-error
        self.machine.events[state].trigger(self)

    def trigger_next(self):
        self.p_prev_state = self.state
        if self.p_next_state:
            self.machine_trigger(self.p_next_state)
            self.p_next_state = None

    def copro_queue_next(self, next_state):
        self.p_next_copro_state = next_state
        self.copro_queue()  # pytype: disable=attribute-error

    def copro_machine_trigger(self, state):
        self.copro_machine.events[state].trigger(
            self)  # pytype: disable=attribute-error

    def copro_trigger_next(self):
        if self.p_next_copro_state:
            self.copro_machine_trigger(self.p_next_copro_state)
            self.p_next_copro_state = None

    def operation_active(self):
        return self.state == 'operating'

    def operation_requested(self, next_state=None):
        if next_state is None:
            next_state = self.p_next_state
        return next_state == 'operate'

    def force_unknown(self):
        self.unknown()  # pytype: disable=attribute-error
        self.p_next_state = None

    def default(self):
        if not self.ignore:
            if self.state != 'unknown':
                if self.state == 'operating':
                    self.p_next_state = 'operate'
                elif self.state == 'queued':
                    self.p_next_state = 'queue'
                elif self.state == 'known':
                    self.p_next_state = self.state
                self.unknown()  # pytype: disable=attribute-error

    @staticmethod
    def make_hash(machine, trunk=False):
        ''' hash the unique metadata parts of an endpoint '''
        h = hashlib.new('ripemd160')
        words = ['tenant', 'mac', 'segment']
        if trunk:
            words.append('ipv4')
            words.append('ipv6')
        pre_h = ''.join([str(machine.get(word, 'missing')) for word in words])
        h.update(pre_h.encode('utf-8'))
        post_h = h.hexdigest()
        return post_h


def endpoint_factory(hashed_val):
    endpoint = Endpoint(hashed_val)
    machine = Machine(
        model=endpoint,
        model_attribute='state',
        states=Endpoint.states,
        transitions=Endpoint.transitions,
        initial='unknown',
        send_event=True)
    machine.name = endpoint.name[:8]+' '
    endpoint.machine = machine
    copro_machine = Machine(
        model=endpoint,
        model_attribute='copro_state',
        states=Endpoint.copro_states,
        transitions=Endpoint.copro_transitions,
        initial='copro_unknown',
        send_event=True)
    copro_machine.name = endpoint.name[:8]+'_copro'
    endpoint.copro_machine = copro_machine
    return endpoint


class EndpointDecoder:

    def __init__(self, endpoint):
        if isinstance(endpoint, dict):
            e = endpoint
        else:
            e = json.loads(endpoint)
        self.endpoint = endpoint_factory(e['name'])
        self.endpoint.state = e['state']
        self.endpoint.copro_state = e.get('copro_state', None)
        self.endpoint.ignore = bool(e.get('ignore', False))
        self.endpoint.metadata = e.get('metadata', {})
        self.endpoint.acl_data = e.get('acl_data', [])
        self.endpoint.endpoint_data = e['endpoint_data']
        self.endpoint.p_next_state = e['p_next_state']
        self.endpoint.p_prev_state = e.get('p_prev_state', None)
        self.endpoint.observed_time = e.get('observed_time', 0)

    def get_endpoint(self):
        return self.endpoint



# -*- coding: utf-8 -*-
"""
Created on 9 December 2018
@author: Charlie Lewis
"""
import ast
import json
import logging

import httpx
from poseidon_core.helpers.config import Config


class Collector(object):

    def __init__(self, endpoint, switch, iterations=1):
        self.logger = logging.getLogger('collector')
        self.config = Config().get_config()
        self.endpoint = endpoint
        self.id = endpoint.name
        self.mac = endpoint.endpoint_data['mac']
        self.nic = None
        nic = self.config['collector_nic']
        try:
            eval_nic = ast.literal_eval(nic)
            if switch in eval_nic:
                self.nic = eval_nic[switch]
            else:
                self.logger.error(
                    'Failed to get collector nic for the switch: {0}'.format(switch))
        except ValueError:
            self.nic = nic
        self.interval = str(self.config['reinvestigation_frequency'])
        self.iterations = str(iterations)

    def start_collector(self):
        '''
        Starts collector for a given endpoint with the
        options passed in at the creation of the class instance.
        '''
        status = False
        payload = {
            'nic': self.nic,
            'id': self.id,
            'interval': self.interval,
            'filter': '\'ether host {0}\''.format(self.mac),
            'iters': self.iterations,
            'metadata': "{'endpoint_data': " + str(self.endpoint.endpoint_data) + '}'}

        self.logger.debug('Payload: {0}'.format(str(payload)))

        network_tap_addr = self.config['network_tap_ip'] + \
            ':' + self.config['network_tap_port']
        uri = 'http://' + network_tap_addr + '/create'

        try:
            resp = httpx.post(uri, json=payload)
            # TODO improve logged output
            self.logger.debug(
                'Collector response: {0}'.format(resp.text))
            response = ast.literal_eval(resp.text)
            if response[0]:
                self.logger.info(
                    'Successfully started the collector for: {0}'.format(self.id))
                self.endpoint.endpoint_data['container_id'] = response[1].rsplit(
                    ':', 1)[-1].strip()
                status = True
            else:
                self.logger.error(
                    'Failed to start collector because: {0}'.format(response[1]))
        except Exception as e:  # pragma: no cover
            self.logger.error(
                'Failed to start collector because: {0}'.format(str(e)))
        return status

    def stop_collector(self):
        '''
        Stops collector for a given endpoint.
        '''
        status = False
        if 'container_id' not in self.endpoint.endpoint_data:
            self.logger.warning(
                'No collector to stop because no container_id for endpoint')
            return True

        payload = {'id': [self.endpoint.endpoint_data['container_id']]}
        self.logger.debug('Payload: {0}'.format(str(payload)))

        network_tap_addr = self.config['network_tap_ip'] + \
            ':' + self.config['network_tap_port']
        uri = 'http://' + network_tap_addr + '/stop'

        try:
            resp = httpx.post(uri, json=payload)
            self.logger.debug(
                'Collector response: {0}'.format(resp.text))
            response = ast.literal_eval(resp.text)
            if response[0]:
                self.logger.info(
                    'Successfully stopped the collector for: {0}'.format(self.id))
                status = True
            else:
                self.logger.error(
                    'Failed to stop collector because response failed with: {0}'.format(response[1]))
        except Exception as e:  # pragma: no cover
            self.logger.error(
                'Failed to stop collector because: {0}'.format(str(e)))
        return status

    # returns a dictionary of existing collectors keyed on dev_hash
    def get_collectors(self):
        network_tap_addr = self.config['network_tap_ip'] + \
            ':' + self.config['network_tap_port']
        uri = 'http://' + network_tap_addr + '/list'
        collectors = {}
        try:
            resp = httpx.get(uri)
            text = resp.text
            # TODO need to parse out text
            self.logger.debug('collector list response: ' + text)
        except Exception as e:  # pragma: no cover
            self.logger.debug(
                'failed to get collector statuses' + str(e))

        return collectors

    def host_has_active_collectors(self, dev_hash):
        active_collectors_exist = False

        collectors = self.get_collectors()

        if dev_hash in collectors:
            hash_coll = collectors[dev_hash]
        else:
            self.logger.warning(
                'Key: {0} not found in collector dictionary. '
                'Treating this as the existence of multiple active'
                'collectors'.format(dev_hash)
            )
            return True

        for c in collectors:
            self.logger.debug(c)
            if (
                collectors[c].hash != dev_hash and
                collectors[c].host == hash_coll.host and
                collectors[c].status != 'exited'
            ):
                active_collectors_exist = True
                break

        return active_collectors_exist



import difflib
import ipaddress
import json
import time
from copy import deepcopy

import pika
from poseidon_core.constants import NO_DATA
from poseidon_core.controllers.faucet.config import FaucetRemoteConfGetSetter
from poseidon_core.controllers.faucet.faucet import FaucetProxy
from poseidon_core.helpers.actions import Actions
from poseidon_core.helpers.endpoint import Endpoint
from poseidon_core.helpers.endpoint import endpoint_factory
from poseidon_core.helpers.endpoint import MACHINE_IP_FIELDS
from poseidon_core.helpers.endpoint import MACHINE_IP_PREFIXES
from poseidon_core.helpers.metadata import DNSResolver
from poseidon_core.helpers.metadata import get_ether_vendor
from poseidon_core.helpers.prometheus import Prometheus


class SDNConnect:

    def __init__(self, config, logger, prom, faucetconfgetsetter_cl=FaucetRemoteConfGetSetter):
        self.config = config
        self.r = None
        self.sdnc = None
        self.endpoints = {}
        self.investigations = 0
        self.coprocessing = 0
        trunk_ports = self.config['trunk_ports']
        if isinstance(trunk_ports, str):
            self.trunk_ports = json.loads(trunk_ports)
        else:
            self.trunk_ports = trunk_ports
        self.logger = logger
        self.prom = prom
        self.faucetconfgetsetter_cl = faucetconfgetsetter_cl
        self.get_sdn_context()
        self.dns_resolver = DNSResolver()
        self.get_stored_endpoints()

    def mirror_endpoint(self, endpoint):
        ''' mirror an endpoint. '''
        status = Actions(endpoint, self.sdnc).mirror_endpoint()
        if not status:
            self.logger.warning(
                'Unable to mirror the endpoint: {0}'.format(endpoint.name))

    def unmirror_endpoint(self, endpoint):
        ''' unmirror an endpoint. '''
        if endpoint.operation_active():
            status = Actions(endpoint, self.sdnc).unmirror_endpoint()
            if not status:
                self.logger.warning(
                    'Unable to unmirror the endpoint: {0}'.format(endpoint.name))
            endpoint.force_unknown()
        else:
            self.logger.info('Not unmirroring endpoint {0} in state {1}'.format(
                endpoint.name, endpoint.state))

    def clear_filters(self):
        ''' clear any exisiting filters. '''
        if isinstance(self.sdnc, FaucetProxy):
            self.sdnc.clear_mirrors()

    def default_endpoints(self):
        ''' set endpoints to default state. '''
        self.clear_filters()
        for endpoint in self.endpoints.values():
            endpoint.default()

    def get_stored_endpoints(self):
        ''' load existing endpoints from Prometheus. '''
        new_endpoints = self.prom.get_stored_endpoints()
        if new_endpoints:
            self.logger.info(
                f'Loaded {len(new_endpoints)} endpoints previously learned.')
            self.endpoints = new_endpoints

    def get_sdn_context(self):
        controller_type = self.config.get('TYPE', None)
        if controller_type == 'faucet':
            self.sdnc = FaucetProxy(
                self.config, faucetconfgetsetter_cl=self.faucetconfgetsetter_cl)
        else:
            self.logger.error(
                'Unknown SDN controller config: {0}'.format(
                    self.config))

    def not_ignored_endpoints(self, state=None):
        endpoints = [endpoint for endpoint in self.endpoints.values()
                     if not endpoint.ignore]
        if state:
            endpoints = [
                endpoint for endpoint in endpoints if endpoint.state == state]
        return endpoints

    def not_copro_ignored_endpoints(self, state=None):
        endpoints = [endpoint for endpoint in self.endpoints.values()
                     if not endpoint.copro_ignore]
        if state:
            endpoints = [
                endpoint for endpoint in endpoints if endpoint.copro_state == state]
        return endpoints

    def endpoint_by_name(self, name):
        return self.endpoints.get(name, None)

    def endpoint_by_hash(self, hash_id):
        return self.endpoint_by_name(hash_id)

    def endpoints_by_ip(self, ip):
        return [
            endpoint for endpoint in self.endpoints.values()
            if ip == endpoint.endpoint_data.get('ipv4', None) or
            ip == endpoint.endpoint_data.get('ipv6', None)]

    def endpoints_by_mac(self, mac):
        return [
            endpoint for endpoint in self.endpoints.values()
            if mac == endpoint.endpoint_data['mac']]

    def investigation_budget(self):
        self.investigations = len([
            endpoint for endpoint in self.not_ignored_endpoints()
            if endpoint.operation_active()])
        return max(
            self.config['max_concurrent_reinvestigations'] - self.investigations, 0)

    def coprocessing_budget(self):
        self.coprocessing = len([
            endpoint for endpoint in self.not_ignored_endpoints()
            if endpoint.copro_state == 'copro_coprocessing'])
        return max(
            self.config['max_concurrent_coprocessing'] - self.coprocessing, 0)

    @staticmethod
    def _connect_rabbit():
        # Rabbit settings
        exchange = 'topic-poseidon-internal'
        exchange_type = 'topic'

        # Starting rabbit connection
        connection = pika.BlockingConnection(
            pika.ConnectionParameters(host='RABBIT_SERVER')
        )

        channel = connection.channel()
        channel.exchange_declare(
            exchange=exchange, exchange_type=exchange_type
        )

        return channel, exchange, connection

    @staticmethod
    def publish_action(action, message):
        try:
            channel, exchange, connection = SDNConnect._connect_rabbit()
            channel.basic_publish(exchange=exchange,
                                  routing_key=action,
                                  body=message)
            connection.close()
        except Exception as e:  # pragma: no cover
            print(str(e))

    def show_endpoints(self, arg):
        endpoints = []
        if arg == 'all':
            endpoints = list(self.endpoints.values())
        else:
            show_type, arg = arg.split(' ', 1)
            for endpoint in self.endpoints.values():
                if show_type == 'state':
                    if arg == 'active':
                        endpoints.append(endpoint)
                    elif arg == 'ignored' and endpoint.ignore:
                        endpoints.append(endpoint)
                    elif endpoint.state == arg:
                        endpoints.append(endpoint)
                elif show_type in ['os', 'role']:
                    mac_addresses = endpoint.metadata.get(
                        'mac_addresses', None)
                    endpoint_mac = endpoint.endpoint_data['mac']
                    if endpoint_mac and mac_addresses and endpoint_mac in mac_addresses:
                        timestamps = mac_addresses[endpoint_mac]
                        try:
                            newest = sorted(
                                [timestamp for timestamp in timestamps])[-1]
                            newest = timestamps[newest]
                        except IndexError:
                            newest = None
                        if newest:
                            if 'labels' in newest:
                                if arg.replace('-', ' ') == newest['labels'][0].lower():
                                    endpoints.append(endpoint)

                    # filter by operating system
                    for ip_field in MACHINE_IP_FIELDS:
                        ip_addresses_field = '_'.join((ip_field, 'addresses'))
                        ip_addresses = endpoint.metadata.get(
                            ip_addresses_field, None)
                        machine_ip = endpoint.endpoint_data.get(ip_field, None)
                        if machine_ip and ip_addresses and machine_ip in ip_addresses:
                            metadata = ip_addresses[machine_ip]
                            os = metadata.get('os', None)
                            if os and os.lower() == arg:
                                endpoints.append(endpoint)
        return endpoints

    def check_endpoints(self, messages):
        if not self.sdnc:
            return

        current = None
        parsed = None

        try:
            current = self.sdnc.get_endpoints(messages=messages)
            parsed = self.sdnc.format_endpoints(current)
        except Exception as e:  # pragma: no cover
            self.logger.error(
                'Could not establish connection to controller because {0}.'.format(e))

        self.find_new_machines(parsed)

    @staticmethod
    def _diff_machine(machine_a, machine_b):

        def _machine_strlines(machine):
            return str(json.dumps(machine, indent=2)).splitlines()

        machine_a_strlines = _machine_strlines(machine_a)
        machine_b_strlines = _machine_strlines(machine_b)
        return '\n'.join(difflib.unified_diff(
            machine_a_strlines, machine_b_strlines, n=1))

    @staticmethod
    def _parse_machine_ip(machine):
        machine_ips = set()
        machine_ip_data = {}
        for ip_field, fields in MACHINE_IP_FIELDS.items():
            try:
                raw_field = machine.get(ip_field, None)
                machine_ip = str(ipaddress.ip_address(raw_field))
                machine_subnet = str(ipaddress.ip_network(machine_ip).supernet(
                    new_prefix=MACHINE_IP_PREFIXES[ip_field]))
            except ValueError:
                machine_ip = None
                machine_subnet = None
            machine_ip_data[ip_field] = ''
            if machine_ip:
                machine_ips.add(machine_ip)
                machine_ip_data.update({
                    ip_field: machine_ip,
                    '_'.join((ip_field, 'subnet')): machine_subnet})
            for field in fields:
                if field not in machine_ip_data:
                    machine_ip_data[field] = NO_DATA
        machine.update(machine_ip_data)
        return machine_ips

    @staticmethod
    def _update_machine_rdns(machine, resolved_machine_ips):
        for ip_field in MACHINE_IP_FIELDS:
            rdns_field = '_'.join((ip_field, 'rdns'))
            machine_ip = machine[ip_field]
            resolved_ip = resolved_machine_ips.get(machine_ip, machine_ip)
            machine[rdns_field] = resolved_ip

    @staticmethod
    def merge_machine_ip(old_machine, new_machine):
        for ip_field, fields in MACHINE_IP_FIELDS.items():
            ip = new_machine.get(ip_field, None)
            old_ip = old_machine.get(ip_field, None)
            if not ip and old_ip:
                new_machine[ip_field] = old_ip
                for field in fields:
                    if field in old_machine:
                        new_machine[field] = old_machine[field]

    def find_new_machines(self, machines):
        '''parse switch structure to find new machines added to network
        since last call'''

        change_acls = False
        machine_ips = set()

        for machine in machines:
            machine['ether_vendor'] = get_ether_vendor(
                machine['mac'], '/poseidon/src/core/core/metadata/nmap-mac-prefixes.txt')
            machine_ips.update(self._parse_machine_ip(machine))
            if 'controller_type' not in machine:
                machine.update({
                    'controller_type': 'none',
                    'controller': ''})

        if machine_ips:
            self.logger.debug('resolving %s' % machine_ips)
            resolved_machine_ips = self.dns_resolver.resolve_ips(
                list(machine_ips))
            self.logger.debug('resolver results %s', resolved_machine_ips)
            for machine in machines:
                self._update_machine_rdns(machine, resolved_machine_ips)

        for machine in machines:
            trunk = False
            for sw in self.trunk_ports:
                if sw == machine['segment'] and self.trunk_ports[sw].split(',')[1] == str(machine['port']) and self.trunk_ports[sw].split(',')[0] == machine['mac']:
                    trunk = True

            h = Endpoint.make_hash(machine, trunk=trunk)
            ep = self.endpoints.get(h, None)
            if ep is None:
                change_acls = True
                m = endpoint_factory(h)
                m.endpoint_data = deepcopy(machine)
                m.touch()
                self.endpoints[m.name] = m
                self.logger.info(
                    'Detected new endpoint: {0}:{1}'.format(m.name, machine))
                continue

            self.merge_machine_ip(ep.endpoint_data, machine)
            if ep.endpoint_data != machine and not ep.ignore:
                diff_txt = self._diff_machine(ep.endpoint_data, machine)
                self.logger.info(
                    'Endpoint changed: {0}:\n{1}'.format(h, diff_txt))
                change_acls = True
                ep.endpoint_data = deepcopy(machine)
            ep.touch()

        if change_acls and self.config['AUTOMATED_ACLS']:
            status = Actions(None, self.sdnc).update_acls(
                rules_file=self.config['RULES_FILE'],
                endpoints=self.endpoints.values())
            if isinstance(status, list):
                self.logger.info(
                    'Automated ACLs did the following: {0}'.format(status[1]))
                for item in status[1]:
                    machine = {'mac': item[1],
                               'segment': item[2], 'port': item[3]}
                    h = Endpoint.make_hash(machine)
                    ep = self.endpoints.get(h, None)
                    if ep:
                        ep.acl_data.append(
                            ((item[0], item[4], item[5]), int(time.time())))

    @staticmethod
    def coprocess_endpoint(_endpoint):
        '''TODO.'''
        return

    @staticmethod
    def uncoprocess_endpoint(_endpoint):
        '''TODO'''
        return



# -*- coding: utf-8 -*-
"""
Created on 5 December 2018
@author: Charlie Lewis
"""
import datetime
import ipaddress
import logging
import re
import time
from collections import defaultdict

import httpx
from poseidon_core import __version__
from poseidon_core.constants import NO_DATA
from poseidon_core.helpers.config import Config
from poseidon_core.helpers.endpoint import Endpoint
from poseidon_core.helpers.endpoint import EndpointDecoder
from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Info
from prometheus_client import start_http_server
from prometheus_client import Summary


class Prometheus():

    def __init__(self):
        self.logger = logging.getLogger('prometheus')
        self.prom_metrics = {}
        self.config = Config().get_config()
        self.prometheus_addr = self.config['prometheus_ip'] + \
            ':' + self.config['prometheus_port']

    def initialize_metrics(self):
        self.prom_metrics['info'] = Info(
            'poseidon_version', 'Info about Poseidon')
        self.prom_metrics['ipv4_table'] = Gauge('poseidon_endpoint_ip_table',
                                                'IP Table',
                                                ['mac',
                                                 'tenant',
                                                 'segment',
                                                 'port',
                                                 'role',
                                                 'ipv4_os',
                                                 'hash_id'])
        self.prom_metrics['roles'] = Gauge('poseidon_endpoint_roles',
                                           'Number of endpoints by role',
                                           ['role'])
        self.prom_metrics['oses'] = Gauge('poseidon_endpoint_oses',
                                          'Number of endpoints by OS',
                                          ['ipv4_os'])
        self.prom_metrics['current_states'] = Gauge('poseidon_endpoint_current_states',
                                                    'Number of endpoints by current state',
                                                    ['current_state'])
        self.prom_metrics['vlans'] = Gauge('poseidon_endpoint_vlans',
                                           'Number of endpoints by VLAN',
                                           ['tenant'])
        self.prom_metrics['port_tenants'] = Gauge('poseidon_endpoint_port_tenants',
                                                  'Number of tenants by port',
                                                  ['port',
                                                   'tenant'])
        self.prom_metrics['port_hosts'] = Gauge('poseidon_endpoint_port_hosts',
                                                'Number of hosts by port',
                                                ['port'])
        self.prom_metrics['last_rabbitmq_routing_key_time'] = Gauge('poseidon_last_rabbitmq_routing_key_time',
                                                                    'Epoch time when last received a RabbitMQ message',
                                                                    ['routing_key'])
        self.prom_metrics['ncapture_count'] = Counter(
            'poseidon_ncapture_count', 'Number of times ncapture ran')
        self.prom_metrics['method_runtime_secs'] = Summary('poseidon_method_runtime_secs',
                                                           'Time spent in Monitor methods',
                                                           ['method'])
        self.prom_metrics['endpoint_role_confidence_top'] = Gauge('poseidon_role_confidence_top',
                                                                  'Confidence of top role prediction',
                                                                  ['mac',
                                                                   'name',
                                                                   'role',
                                                                   'pcap_labels',
                                                                   'ipv4_os',
                                                                   'ipv4_address',
                                                                   'ipv6_address',
                                                                   'hash_id'])
        self.prom_metrics['endpoint_role_confidence_second'] = Gauge('poseidon_role_confidence_second',
                                                                     'Confidence of second role prediction',
                                                                     ['mac',
                                                                      'name',
                                                                      'role',
                                                                      'pcap_labels',
                                                                      'ipv4_os',
                                                                      'ipv4_address',
                                                                      'ipv6_address',
                                                                      'hash_id'])
        self.prom_metrics['endpoint_role_confidence_third'] = Gauge('poseidon_role_confidence_third',
                                                                    'Confidence of third role prediction',
                                                                    ['mac',
                                                                     'name',
                                                                     'role',
                                                                     'pcap_labels',
                                                                     'ipv4_os',
                                                                     'ipv4_address',
                                                                     'ipv6_address',
                                                                     'hash_id'])
        self.prom_metrics['endpoints'] = Gauge('poseidon_endpoints',
                                               'All endpoints',
                                               ['mac',
                                                'tenant',
                                                'segment',
                                                'ether_vendor',
                                                'controller_type',
                                                'controller',
                                                'name',
                                                'port',
                                                'hash_id'])
        self.prom_metrics['endpoint_state'] = Gauge('poseidon_endpoint_state',
                                                    'State for all endpoints',
                                                    ['mac',
                                                     'tenant',
                                                     'segment',
                                                     'ether_vendor',
                                                     'name',
                                                     'port',
                                                     'state',
                                                     'hash_id'])
        self.prom_metrics['endpoint_os'] = Gauge('poseidon_endpoint_os',
                                                 'Operating System for all endpoints',
                                                 ['mac',
                                                  'tenant',
                                                  'segment',
                                                  'ether_vendor',
                                                  'name',
                                                  'port',
                                                  'ipv4_os',
                                                  'hash_id'])
        self.prom_metrics['endpoint_role'] = Gauge('poseidon_endpoint_role',
                                                   'Top role for all endpoints',
                                                   ['mac',
                                                    'tenant',
                                                    'segment',
                                                    'ether_vendor',
                                                    'name',
                                                    'port',
                                                    'top_role',
                                                    'hash_id'])
        self.prom_metrics['endpoint_ip'] = Gauge('poseidon_endpoint_ip',
                                                 'IP Address for all endpoints',
                                                 ['mac',
                                                  'tenant',
                                                  'segment',
                                                  'ether_vendor',
                                                  'name',
                                                  'port',
                                                  'ipv4_address',
                                                  'ipv6_address',
                                                  'ipv4_subnet',
                                                  'ipv6_subnet',
                                                  'ipv4_rdns',
                                                  'ipv6_rdns',
                                                  'hash_id'])
        self.prom_metrics['endpoint_metadata'] = Gauge('poseidon_endpoint_metadata',
                                                       'Metadata for all endpoints',
                                                       ['mac',
                                                        'tenant',
                                                        'segment',
                                                        'ether_vendor',
                                                        'prev_state',
                                                        'next_state',
                                                        'acls',
                                                        'ignore',
                                                        'ipv4_subnet',
                                                        'ipv6_subnet',
                                                        'ipv4_rdns',
                                                        'ipv6_rdns',
                                                        'controller_type',
                                                        'controller',
                                                        'name',
                                                        'state',
                                                        'port',
                                                        'top_role',
                                                        'ipv4_os',
                                                        'ipv4_address',
                                                        'ipv6_address',
                                                        'hash_id'])

    @staticmethod
    def get_metrics():
        metrics = {'info': defaultdict(int),
                   'roles': defaultdict(int),
                   'oses': defaultdict(int),
                   'current_states': defaultdict(int),
                   'vlans': defaultdict(int),
                   'port_tenants': defaultdict(int),
                   'port_hosts': defaultdict(int),
                   'ncapture_count': 0}
        return metrics

    def update_metrics(self, hosts):

        metrics = Prometheus.get_metrics()
        metrics['info']['version'] = __version__

        for host in hosts:
            metrics['roles'][host['role']] += 1
            metrics['oses'][host['ipv4_os']] += 1
            metrics['vlans'][host['tenant']] += 1
            metrics['port_hosts'][host['port']] += 1
            metrics['port_tenants'][(host['port'], host['tenant'])] += 1
            metrics['current_states'][host['state']] += 1

        if self.prom_metrics:
            for host in hosts:
                if host['ipv4']:
                    self.prom_metrics['ipv4_table'].labels(mac=host['mac'],
                                                           tenant=host['tenant'],
                                                           segment=host['segment'],
                                                           port=host['port'],
                                                           role=host['role'],
                                                           ipv4_os=host['ipv4_os'],
                                                           hash_id=host['id']).set(int(ipaddress.ip_address(host['ipv4'])))
            for role in metrics['roles']:
                self.prom_metrics['roles'].labels(
                    role=role).set(metrics['roles'][role])
            for os_t in metrics['oses']:
                self.prom_metrics['oses'].labels(
                    ipv4_os=os_t).set(metrics['oses'][os_t])
            for current_state in metrics['current_states']:
                self.prom_metrics['current_states'].labels(current_state=current_state).set(
                    metrics['current_states'][current_state])
            for vlan in metrics['vlans']:
                self.prom_metrics['vlans'].labels(
                    tenant=vlan).set(metrics['vlans'][vlan])
            for port_tenant in metrics['port_tenants']:
                self.prom_metrics['port_tenants'].labels(port=port_tenant[0],
                                                         tenant=port_tenant[1]).set(metrics['port_tenants'][port_tenant])
            for port_host in metrics['port_hosts']:
                self.prom_metrics['port_hosts'].labels(
                    port=port_host).set(metrics['port_hosts'][port_host])
            self.prom_metrics['info'].info(metrics['info'])

    @staticmethod
    def latest_metric(metric):
        return metric['values'][-1]

    def latest_value(self, metric):
        return float(self.latest_metric(metric)[1])

    def latest_timestamp(self, metric):
        return self.latest_metric(metric)[0]

    @staticmethod
    def metric_label(metric, label, default_value=NO_DATA):
        return metric['metric'].get(label, default_value)

    def sorted_metrics(self, response):
        ''' return timeseries in order, most recently asserted, first. '''
        return sorted([
            {'metric': result['metric'], 'values': [
                self.latest_metric(result)]}
            for result in response.json()['data']['result']],
            key=lambda x: self.latest_value(x), reverse=True)

    def prom_query(self, var, start_time_str, end_time_str, step='30s'):
        payload = {'query': var, 'start': start_time_str,
                   'end': end_time_str, 'step': step}
        try:
            response = httpx.get(
                'http://%s/api/v1/query_range' % self.prometheus_addr, params=payload)
        except Exception:
            return []
        return self.sorted_metrics(response)

    def consolidate_prom(self, mr, r1, r2, r3):
        hashes = {}
        if mr:
            for metric in mr:
                hash_id = self.metric_label(metric, 'hash_id')
                if hash_id not in hashes:
                    hashes[hash_id] = metric['metric']
        role_hashes = {}
        if r1:
            for metric in r1:
                hash_id = self.metric_label(metric, 'hash_id')
                if not hash_id in role_hashes:
                    role_hashes[hash_id] = {
                        'mac': self.metric_label(metric, 'mac'),
                        'top_role': self.metric_label(metric, 'role'),
                        'top_confidence': self.latest_value(metric),
                        'pcap_labels': self.metric_label(metric, 'pcap_labels', '')}
        if r2:
            for metric in r2:
                hash_id = self.metric_label(metric, 'hash_id')
                if hash_id in role_hashes:
                    role_hashes[hash_id].update({
                        'second_role': self.metric_label(metric, 'role'),
                        'second_confidence': self.latest_value(metric)})
        if r3:
            for metric in r3:
                hash_id = self.metric_label(metric, 'hash_id')
                if hash_id in role_hashes:
                    role_hashes[hash_id].update({
                        'third_role': self.metric_label(metric, 'role'),
                        'third_confidence': self.latest_value(metric)})

        return hashes, role_hashes

    def scrape_prom(self):
        current_time = datetime.datetime.utcnow()
        # 6 hours in the past and 2 hours in the future
        start_time = current_time - datetime.timedelta(hours=6)
        end_time = current_time + datetime.timedelta(hours=2)
        start_time_str = start_time.isoformat()[:-4]+'Z'
        end_time_str = end_time.isoformat()[:-4]+'Z'
        mr = self.prom_query('poseidon_endpoint_metadata',
                             start_time_str, end_time_str)
        r1 = self.prom_query(
            'poseidon_role_confidence_top{role!="%s"}' % NO_DATA, start_time_str, end_time_str)
        r2 = self.prom_query(
            'poseidon_role_confidence_second{role!="%s"}' % NO_DATA, start_time_str, end_time_str)
        r3 = self.prom_query(
            'poseidon_role_confidence_third{role!="%s"} % NO_DATA', start_time_str, end_time_str)
        return self.consolidate_prom(mr, r1, r2, r3)

    @staticmethod
    def prom_endpoints(hashes, role_hashes):
        endpoints = {}
        for p_endpoint in hashes.values():
            prev_state = p_endpoint.get('prev_state', None)
            if prev_state not in Endpoint.transitions:
                prev_state = None
            next_state = p_endpoint.get('next_state', None)
            if next_state not in Endpoint.transitions:
                next_state = None
            state = p_endpoint['state']
            p_endpoint.update({
                'name': p_endpoint['hash_id'],
                'p_next_state': next_state,
                'p_prev_state': prev_state,
                'acl_data': [],  # TODO: acl_data
                'metadata': {'mac_addresses': {}, 'ipv4_addresses': {}, 'ipv6_addresses': {}},
                'state': state,
                'ignore': False,  # TODO: force ignore off
                'endpoint_data': {
                    'mac': p_endpoint['mac'],
                    'segment': p_endpoint['segment'],
                    'port': p_endpoint['port'],
                    'vlan': p_endpoint['tenant'],
                    'tenant': p_endpoint['tenant'],
                    'ipv4': p_endpoint.get('ipv4_address', ''),
                    'ipv6': p_endpoint.get('ipv6_address', ''),
                    'controller_type': p_endpoint['controller_type'],
                    'controller': p_endpoint.get('controller', ''),
                    'name': p_endpoint['name'],
                    'ether_vendor': p_endpoint['ether_vendor'],
                    'ipv4_subnet': p_endpoint.get('ipv4_subnet', ''),
                    'ipv4_rdns': p_endpoint.get('ipv4_rdns', ''),
                    'ipv6_rdns': p_endpoint.get('ipv6_rdns', ''),
                    'ipv6_subnet': p_endpoint.get('ipv6_subnet', '')}})
            ipv4 = p_endpoint.get('ipv4_address', '')
            ipv4_os = p_endpoint.get('ipv4_os', '')
            if ipv4 and ipv4_os:
                p_endpoint['metadata']['ipv4_addresses'][ipv4] = {
                    'short_os': ipv4_os}
            mac = p_endpoint['mac']
            for role_hash in role_hashes.values():
                role_mac = role_hash['mac']
                if mac != role_mac:
                    continue
                if not mac in p_endpoint['metadata']['mac_addresses']:
                    roles = [
                        role_hash.get('top_role', NO_DATA),
                        role_hash.get('second_role', NO_DATA),
                        role_hash.get('third_role', NO_DATA)]
                    confidences = [
                        role_hash.get('top_confidence', NO_DATA),
                        role_hash.get('second_confidence', NO_DATA),
                        role_hash.get('third_confidence', NO_DATA)]
                    pcap_labels = role_hash['pcap_labels']
                    p_endpoint['metadata']['mac_addresses'][mac] = {
                        'classification': {
                            'labels': roles,
                            'confidences': confidences,
                        },
                        'pcap_labels': pcap_labels}
            endpoint = EndpointDecoder(p_endpoint).get_endpoint()
            endpoints[endpoint.name] = endpoint
        return endpoints

    def get_stored_endpoints(self):
        ''' load existing endpoints from Prometheus. '''
        hashes, role_hashes = self.scrape_prom()
        return self.prom_endpoints(hashes, role_hashes)

    def runtime_callable(self, method):
        method_name = str(method)
        method_re = re.compile(r'.+bound method (\S+).+')
        method_match = method_re.match(method_name)
        if method_match:
            method_name = method_match.group(1)
        with self.prom_metrics['method_runtime_secs'].labels(method=method_name).time():
            return method()

    def update_endpoint_metadata(self, endpoints):
        update_time = time.time()
        for hash_id, endpoint in endpoints.items():
            ipv4 = endpoint.endpoint_data['ipv4']
            ipv6 = endpoint.endpoint_data['ipv6']
            ipv4_subnet = endpoint.endpoint_data['ipv4_subnet']
            ipv6_subnet = endpoint.endpoint_data['ipv6_subnet']
            ipv4_rdns = endpoint.endpoint_data['ipv4_rdns']
            ipv6_rdns = endpoint.endpoint_data['ipv6_rdns']
            port = endpoint.endpoint_data['port']
            tenant = endpoint.endpoint_data['tenant']
            segment = endpoint.endpoint_data['segment']
            ether_vendor = endpoint.endpoint_data['ether_vendor']
            controller = endpoint.endpoint_data['controller']
            controller_type = endpoint.endpoint_data['controller_type']
            roles, confidences, pcap_labels = endpoint.get_roles_confidences_pcap_labels()
            top_role, second_role, third_role = roles
            top_conf, second_conf, third_conf = confidences
            ipv4_os = endpoint.get_ipv4_os()

            def set_prom(var, val, **prom_labels):
                prom_labels.update({
                    'mac': endpoint.endpoint_data['mac'],
                    'name': endpoint.endpoint_data['name'],
                    'hash_id': hash_id,
                })
                try:
                    self.prom_metrics[var].labels(**prom_labels).set(val)
                except ValueError:
                    pass

            def set_prom_role(var, val, role):
                set_prom(
                    var,
                    val,
                    role=role,
                    ipv4_os=ipv4_os,
                    ipv4_address=ipv4,
                    ipv6_address=ipv6,
                    pcap_labels=pcap_labels)

            def update_prom(var, **prom_labels):
                prom_labels.update({
                    'tenant': tenant,
                    'segment': segment,
                    'ether_vendor': ether_vendor,
                    'port': port,
                })
                set_prom(var, update_time, **prom_labels)

            set_prom_role(
                'endpoint_role_confidence_top',
                top_conf,
                top_role)
            set_prom_role(
                'endpoint_role_confidence_second',
                second_conf,
                second_role)
            set_prom_role(
                'endpoint_role_confidence_third',
                third_conf,
                third_role)
            update_prom(
                'endpoints',
                controller_type=controller_type,
                controller=controller)
            update_prom(
                'endpoint_state',
                state=endpoint.state)
            update_prom(
                'endpoint_os',
                ipv4_os=ipv4_os)
            update_prom(
                'endpoint_role',
                top_role=top_role)
            update_prom(
                'endpoint_ip',
                ipv4_subnet=ipv4_subnet,
                ipv6_subnet=ipv6_subnet,
                ipv4_rdns=ipv4_rdns,
                ipv6_rdns=ipv6_rdns,
                ipv4_address=ipv4,
                ipv6_address=ipv6)
            update_prom(
                'endpoint_metadata',
                prev_state=endpoint.p_prev_state,
                next_state=endpoint.p_next_state,
                acls=endpoint.acl_data,
                ignore=str(endpoint.ignore),
                ipv4_subnet=ipv4_subnet,
                ipv6_subnet=ipv6_subnet,
                ipv4_rdns=ipv4_rdns,
                ipv6_rdns=ipv6_rdns,
                controller_type=controller_type,
                controller=controller,
                state=endpoint.state,
                top_role=top_role,
                ipv4_os=ipv4_os,
                ipv4_address=ipv4,
                ipv6_address=ipv6)

    @staticmethod
    def start(port=9304):
        start_http_server(port)



# -*- coding: utf-8 -*-
"""
Created on 19 February 2019
@author: Charlie Lewis
"""
import functools
import socket
from concurrent.futures import ThreadPoolExecutor

from poseidon_core.constants import NO_DATA


@functools.lru_cache()
def get_ether_vendor(mac, lookup_path):
    """
    Takes a MAC address and looks up and returns the vendor for it.
    """
    mac = ''.join(mac.split(':'))[:6].upper()
    try:
        with open(lookup_path, 'r') as f:
            for line in f:
                if line.startswith(mac):
                    return line.split()[1].strip()
    except Exception:  # pragma: no cover
        return NO_DATA


class DNSResolver:

    TIMEOUT = 5

    @staticmethod
    def _resolve_ip(ip):
        try:
            result = socket.getnameinfo((ip, 0), 0)[0]
            if result == ip:
                return NO_DATA
            return result
        except socket.gaierror:
            return NO_DATA

    def resolve_ips(self, ips):
        with ThreadPoolExecutor() as executor:
            return {ip: result for ip, result in zip(ips, executor.map(DNSResolver()._resolve_ip, list(ips)))}



# -*- coding: utf-8 -*-
"""
Created on 5 December 2018
@author: Charlie Lewis
"""
import configparser
import json
import logging
import os
import tempfile
from distutils import util

import yaml
from poseidon_core.helpers.exception_decor import exception


class Config():

    def __init__(self):
        self.logger = logging.getLogger('config')
        self.config = configparser.RawConfigParser()
        self.config.optionxform = str
        if os.environ.get('POSEIDON_CONFIG') is not None:
            self.config_path = os.environ.get('POSEIDON_CONFIG')
        else:  # pragma: no cover
            raise Exception(
                'Could not find poseidon config. Make sure to set the POSEIDON_CONFIG environment variable')
        self.config.read_file(open(self.config_path, 'r'))

    def get_config(self):
        # set some defaults
        controller = {
            'TYPE': 'faucet',
            'RULES_FILE': None,
            'MIRROR_PORTS': None,
            'AUTOMATED_ACLS': False,
            'LEARN_PUBLIC_ADDRESSES': False,
            'reinvestigation_frequency': 900,
            'max_concurrent_reinvestigations': 2,
            'max_concurrent_coprocessing': 2,
            'logger_level': 'INFO',
            'faucetconfrpc_address': 'faucetconfrpc:59999',
            'faucetconfrpc_client': 'faucetconfrpc',
            'prometheus_ip': 'prometheus',
            'prometheus_port': 9090,
        }

        config_map = {
            'controller_type': ('TYPE', []),
            'learn_public_addresses': ('LEARN_PUBLIC_ADDRESSES', [util.strtobool]),
            'rules_file': ('RULES_FILE', []),
            'collector_nic': ('collector_nic', []),
            'controller_mirror_ports': ('MIRROR_PORTS', [json.loads]),
            'controller_proxy_mirror_ports': ('controller_proxy_mirror_ports', [json.loads]),
            'tunnel_vlan': ('tunnel_vlan', [int]),
            'tunnel_name': ('tunnel_name', []),
            'automated_acls': ('AUTOMATED_ACLS', [util.strtobool]),
            'FA_RABBIT_PORT': ('FA_RABBIT_PORT', [int]),
            'scan_frequency': ('scan_frequency', [int]),
            'reinvestigation_frequency': ('reinvestigation_frequency', [int]),
            'max_concurrent_reinvestigations': ('max_concurrent_reinvestigations', [int]),
            'max_concurrent_coprocessing': ('max_concurrent_coprocessing', [int]),
            'ignore_vlans': ('ignore_vlans', [json.loads]),
            'ignore_ports': ('ignore_ports', [json.loads]),
            'trunk_ports': ('trunk_ports', [json.loads]),
            'logger_level': ('logger_level', []),
        }

        for section in self.config.sections():
            for key, val in self.config[section].items():
                if isinstance(val, str):
                    val = val.strip("'")
                controller_key, val_funcs = config_map.get(key, (key, []))
                for val_func in val_funcs:
                    try:
                        val = val_func(val)
                        break
                    except Exception as e:  # pragma: no cover
                        self.logger.error(
                            'Unable to set configuration option {0} {1} because {2}'.format(key, val, str(e)))
                controller[controller_key] = val
        return controller


def represent_none(dumper, _):
    return dumper.represent_scalar('tag:yaml.org,2002:null', '')


def parse_rules(config_file):
    obj_doc = yaml_in(config_file)
    return obj_doc


@exception
def yaml_in(config_file):
    try:
        with open(config_file, 'r') as stream:
            return yaml.safe_load(stream)
    except Exception as e:  # pragma: no cover
        return False


@exception
def yaml_out(config_file, obj_doc):
    stream = tempfile.NamedTemporaryFile(
        prefix=os.path.basename(config_file),
        dir=os.path.dirname(config_file),
        mode='w',
        delete=False)
    yaml.add_representer(type(None), represent_none)
    yaml.dump(obj_doc, stream, default_flow_style=False)
    stream.close()
    os.replace(stream.name, config_file)
    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json
import queue
import random
import sys
import time


class Monitor:

    def __init__(self, logger, config, schedule, job_queue, sdnc, prom):
        self.logger = logger
        self.rabbits = []
        self.config = config
        self.job_queue = job_queue
        self.sdnc = sdnc
        self.prom = prom

        # timer class to call things periodically in own thread
        schedule.every(self.config['scan_frequency']).seconds.do(
            self.schedule_job_update_metrics)
        schedule.every(self.config['reinvestigation_frequency']).seconds.do(
            self.schedule_job_reinvestigation_timeout)

    def get_hosts(self):
        # TODO consolidate with update_endpoint_metadata
        hosts = []
        for hash_id, endpoint in self.sdnc.endpoints.items():
            roles, _, _ = endpoint.get_roles_confidences_pcap_labels()
            role = roles[0]
            ipv4_os = endpoint.get_ipv4_os()
            host = {
                'mac': endpoint.endpoint_data['mac'],
                'id': hash_id,
                'role': role,
                'ipv4_os': ipv4_os,
                'state': endpoint.state,
                'tenant': endpoint.endpoint_data['tenant'],
                'port': endpoint.endpoint_data['port'],
                'segment': endpoint.endpoint_data['segment'],
                'ipv4': endpoint.endpoint_data['ipv4']}
            hosts.append(host)
        return hosts

    def job_update_metrics(self):
        self.logger.debug('updating metrics')
        try:
            hosts = self.get_hosts()
            self.prom.update_metrics(hosts)
        except Exception as e:  # pragma: no cover
            self.logger.error(
                'Unable to get current state and send it to Prometheus because: {0}'.format(str(e)))
        return 0

    def job_recoprocess(self):
        if not self.sdnc.sdnc:
            for endpoint in self.sdnc.not_copro_ignored_endpoints():
                if endpoint.copro_state != 'copro_nominal':
                    endpoint.copro_nominal()  # pytype: disable=attribute-error
            return 0
        events = 0
        for endpoint in self.sdnc.not_copro_ignored_endpoints('copro_coprocessing'):
            if endpoint.copro_state_timeout(2*self.config['coprocessing_frequency']):
                self.logger.debug(
                    'timing out: {0} and setting to unknown'.format(endpoint.name))
                self.sdnc.uncoprocess_endpoint(endpoint)
                endpoint.copro_unknown()  # pytype: disable=attribute-error
                events += 1
        return events

    def job_reinvestigation_timeout(self):
        ''' put endpoints into the reinvestigation state if possible, and timeout investigations '''
        if not self.sdnc.sdnc:
            for endpoint in self.sdnc.not_ignored_endpoints():
                if endpoint.state != 'known':
                    endpoint.known()  # pytype: disable=attribute-error
            return 0
        events = 0
        timeout = 2*self.config['reinvestigation_frequency']
        for endpoint in self.sdnc.not_ignored_endpoints():
            if endpoint.observed_timeout(timeout):
                self.logger.info(
                    'observation timing out: {0}'.format(endpoint.name))
                endpoint.force_unknown()
                events += 1
            elif endpoint.operation_active() and endpoint.state_timeout(timeout):
                self.logger.info(
                    'mirror timing out: {0}'.format(endpoint.name))
                self.sdnc.unmirror_endpoint(endpoint)
                events += 1
        budget = self.sdnc.investigation_budget()
        candidates = self.sdnc.not_ignored_endpoints('queued')
        if not candidates:
            candidates = self.sdnc.not_ignored_endpoints('known')
        return events + self._schedule_queued_work(
            candidates, budget, 'operate', self.sdnc.mirror_endpoint, shuffle=True)

    def schedule_job_update_metrics(self):
        self.job_queue.put(self.job_update_metrics)

    def schedule_job_reinvestigation_timeout(self):
        self.job_queue.put(self.job_reinvestigation_timeout)

    def _schedule_queued_work(self, queued_endpoints, budget, endpoint_state, endpoint_work, shuffle=False):
        events = 0
        if self.sdnc.sdnc:
            if shuffle:
                random.shuffle(queued_endpoints)
            for endpoint in queued_endpoints[:budget]:
                getattr(endpoint, endpoint_state)()
                endpoint_work(endpoint)
                if endpoint_state in ['trigger_next', 'operate']:
                    # TODO this may not be necessarily true going forward
                    self.prom.prom_metrics['ncapture_count'].inc()
                events += 1
        return events

    # TODO make generic
    def schedule_mirroring(self):
        for endpoint in self.sdnc.not_ignored_endpoints('unknown'):
            endpoint.queue_next('operate')
        budget = self.sdnc.investigation_budget()
        queued_endpoints = [
            endpoint for endpoint in self.sdnc.not_ignored_endpoints('queued')
            if endpoint.operation_requested()]
        queued_endpoints = sorted(queued_endpoints, key=lambda x: x.state_time)
        self.logger.debug('operations {0}, budget {1}, queued {2}'.format(
            str(self.sdnc.investigations), str(budget), str(len(queued_endpoints))))
        return self._schedule_queued_work(queued_endpoints, budget, 'trigger_next', self.sdnc.mirror_endpoint)

    # TODO make generic
    def schedule_coprocessing(self):
        for endpoint in self.sdnc.not_copro_ignored_endpoints('copro_unknown'):
            endpoint.copro_queue_next('copro_coprocess')
        budget = self.sdnc.coprocessing_budget()
        queued_endpoints = self.sdnc.not_copro_ignored_endpoints(
            'copro_queued')
        queued_endpoints = sorted(
            queued_endpoints, key=lambda x: x.copro_state_time)
        self.logger.debug('coprocessing {0}, budget {1}, queued {2}'.format(
            str(self.sdnc.coprocessing), str(budget), str(len(queued_endpoints))))
        return self._schedule_queued_work(queued_endpoints, budget, 'copro_trigger_next', self.sdnc.coprocess_endpoint)



# -*- coding: utf-8 -*-
"""
Created on 31 January 2020
@author: Ryan Ashley
"""
import logging
import os
from pathlib import Path

# TODO merge with primitives/acl.py


class Acl:

    def __init__(self, acl_file=None, faucetconfgetsetter=None):
        self.acls = {}
        self.acl_file = acl_file
        self.faucetconfgetsetter = faucetconfgetsetter

    def _read_existing(self):
        try:
            return self.faucetconfgetsetter.read_faucet_conf(self.acl_file)
        except (FileNotFoundError, PermissionError):
            return {}

    def read(self, config_yaml=None):
        if config_yaml is None:
            config_yaml = self._read_existing()
        self.acls = config_yaml.get('acls', {})
        return config_yaml

    def write(self):
        config_yaml = self._read_existing()
        self._merge_acls(config_yaml)
        try:
            self.faucetconfgetsetter.write_faucet_conf(
                self.acl_file, config_yaml)
            return True
        except (FileNotFoundError, PermissionError):
            return False

    def add_rule(self, name, rule):
        if name not in self.acls:
            self.acls[name] = []
        self.acls[name].append(rule)

    def _merge_acls(self, yaml_config):
        if 'acls' not in yaml_config:
            yaml_config['acls'] = {}
        yaml_config['acls'].update(self.acls)


class ExclusiveAcl(Acl):

    def _merge_acls(self, yaml_config):
        yaml_config['acls'] = self.acls


class VolosAcl(ExclusiveAcl):

    def __init__(self, endpoint, acl_dir, copro_vlans=[2], copro_port=23):
        self.mac = endpoint.endpoint_data['mac']
        self.acl_key = f'volos_copro_{self.mac}'
        self.acl_dir = acl_dir
        acl_file = os.path.join(self.acl_dir, f'/%s.yaml' % self.acl_key)
        super(VolosAcl, self).__init__(acl_file=acl_file)
        self.logger = logging.getLogger('coprocessor')
        self.endpoint = endpoint
        self.id = endpoint.name
        self.copro_vlans = copro_vlans
        self.copro_port = copro_port

    def write_acl_file(self, port_list=[]):
        self.acls = {}
        for port in port_list:
            for eth_type in (0x0800, 0x86dd):
                ip_str = port['proto']
                addresses = self.endpoint.metadata.get(
                    '%s_addresses' % ip_str, None)
                if addresses:
                    for ip in addresses:
                        rule = {'rule': {
                            'dl_type': eth_type,
                            'nw_proto': port['proto_id'],
                            '%s_src' % ip_str: ip,
                            'actions': {
                                'output': {
                                    'ports': [self.copro_port],
                                    'vlan_vid': self.copro_vlans}}}}
                    rule['rule']['%s_dst' % ip_str] = port['port']
                    self.add_rule(self.acl_key, rule)
        self.add_rule(self.acl_key, {'rule': {'actions': {'allow': 1}}})
        status = self.write()
        if not status:
            self.logger.error(
                'Volos ACL file:{0} could not be written. Coprocessing may not work as expected'.format(self.acl_file))
        return status

    def delete_acl_file(self):
        try:
            if os.path.exists(self.acl_file):
                os.remove(self.acl_file)
                return True
        except Exception as e:  # pragma: no cover
            self.logger.error(
                'Volos ACL file:{0} could not be deleted. Coprocessing may not work as expected'.format(self.acl_file))
        return False

    def ensure_acls_dir(self):
        try:
            if not os.path.exists(self.acl_dir):
                Path(self.acl_dir).mkdir(parents=True, exist_ok=True)
                return True
        except Exception as e:  # pragma: no cover
            self.logger.error(
                'Volos ACL directory:{0} could not be created. Coprocessing may not work as expected'.format(self.acl_file))
        return False



# -*- coding: utf-8 -*-
"""
Created on 30 January 2020
@author: Ryan Ashley
"""
import logging


class Coprocess:

    def __init__(self, controller):
        self.logger = logging.getLogger('coprocessor')
        self.pipette_repo = controller['pipette_repo']
        self.pipette_dir = controller['pipette_dir']
        self.coprocessor_nic = controller['coprocessor_nic']
        self.coprocessor_port = controller['coprocessor_port']
        self.coprocessor_vlans = controller['coprocessor_vlans']
        self.fake_interface = controller['fake_interface']
        self.fake_mac = controller['fake_mac']
        self.fake_ips = controller['fake_ips']
        self.bridge = controller['bridge']
        self.pipette_port = controller['pipette_port']
        self.pcap_location = controller['pcap_location']
        self.pcap_size = controller['pcap_size']
        self.pipette_running = False

    def start_coprocessor(self):
        '''
        Starts OVS and containers
        '''
        status = False

        return status

    def stop_coprocessor(self):
        '''
        Stops OVS and containers
        '''
        status = False

        return status



#!/usr/bin/env python
from setuptools import setup

setup(
    name='poseidon_core',
    setup_requires=['pbr>=1.9', 'setuptools>=17.1'],
    python_requires='>=3.6',
    packages=['poseidon_core'],
    package_dir={'poseidon_core': 'core'},
    pbr=True
)



# -*- coding: utf-8 -*-
"""
Created on 21 August 2017
@author: dgrossman
"""
import logging
import threading
import time
from functools import partial

import pika


class Rabbit:
    '''
    Base Class for RabbitMQ
    '''

    def __init__(self):
        self.logger = logging.getLogger('rabbit')
        self.connection = None
        self.channel = None
        self.mq_recv_thread = None
        self.queue_name = 'poseidon_main'

    def close(self):
        if self.connection:
            self.connection.close()

    def make_rabbit_connection(self, host, port, exchange, keys,
                               total_sleep=float('inf')):  # pragma: no cover
        '''
        Connects to rabbitmq using the given hostname,
        exchange, and queue. Retries on failure until success.
        Binds routing keys appropriate for module, and returns
        the channel and connection.
        '''
        wait = True
        do_rabbit = True

        while wait and total_sleep > 0:
            try:
                # Starting rabbit connection
                self.connection = pika.BlockingConnection(
                    pika.ConnectionParameters(host=host, port=port)
                )
                self.channel = self.connection.channel()
                self.channel.exchange_declare(
                    exchange=exchange, exchange_type='topic')
                self.channel.queue_declare(
                    queue=self.queue_name, exclusive=False, durable=True)
                self.logger.info(f'Connected to {host} rabbitmq...')
                wait = False
            except Exception as e:
                self.logger.debug(
                    f'Waiting for connection to {host} rabbitmq...')
                time.sleep(2)
                total_sleep -= 2
                wait = True

        if wait:
            do_rabbit = False

        if self.channel is not None and isinstance(keys, list) and not wait:
            for key in keys:
                self.logger.debug(
                    f'Array adding key:{key} to rabbitmq channel')
                self.channel.queue_bind(
                    exchange=exchange, queue=self.queue_name, routing_key=key)

        if isinstance(keys, str) and not wait:
            self.logger.debug(
                f'String adding key:{keys} to rabbitmq channel')
            self.channel.queue_bind(
                exchange=exchange, queue=self.queue_name, routing_key=keys)

        return do_rabbit

    def start_channel(self, mycallback, m_queue):
        ''' Handle threading for messagetype '''
        self.logger.debug(f'About to start channel {self.channel}')
        self.channel.basic_consume(
            self.queue_name, partial(mycallback, q=m_queue))
        self.mq_recv_thread = threading.Thread(
            target=self.channel.start_consuming)
        self.mq_recv_thread.start()



from pbr.version import VersionInfo

# TODO: 3.8+ and later, use importlib: https://pypi.org/project/importlib-metadata/
__version__ = VersionInfo('poseidon_core').semantic_version().release_string()



import json
import queue
import time
from collections import defaultdict
from functools import partial

from poseidon_core.helpers.actions import Actions
from poseidon_core.helpers.config import Config
from poseidon_core.helpers.rabbit import Rabbit


class SDNEvents:

    def __init__(self, logger, prom, sdnc):
        self.logger = logger
        self.prom = prom
        self.m_queue = queue.Queue()
        self.job_queue = queue.Queue()
        self.rabbits = []
        self.config = Config().get_config()
        self.sdnc = sdnc
        self.sdnc.default_endpoints()
        self.prom.update_endpoint_metadata(self.sdnc.endpoints)

    def create_message_queue(self, host, port, exchange, binding_key):
        waiting = True
        while waiting:
            rabbit = Rabbit()
            rabbit.make_rabbit_connection(
                host, port, exchange, binding_key)
            rabbit.start_channel(
                self.rabbit_callback, self.m_queue)
            waiting = False
        self.rabbits.append(rabbit)

    def start_message_queues(self):
        host = self.config['FA_RABBIT_HOST']
        port = int(self.config['FA_RABBIT_PORT'])
        exchange = 'topic-poseidon-internal'
        binding_key = ['poseidon.algos.#', 'poseidon.action.#']
        self.create_message_queue(
            host, port, exchange, binding_key)
        exchange = self.config['FA_RABBIT_EXCHANGE']
        binding_key = [self.config['FA_RABBIT_ROUTING_KEY']+'.#']
        self.create_message_queue(
            host, port, exchange, binding_key)

    def merge_metadata(self, new_metadata):
        updated = set()
        metadata_types = {
            'mac_addresses': self.sdnc.endpoints_by_mac,
            'ipv4_addresses': self.sdnc.endpoints_by_ip,
            'ipv6_addresses': self.sdnc.endpoints_by_ip,
        }
        for metadata_type, metadata_lookup in metadata_types.items():
            type_new_metadata = new_metadata.get(metadata_type, {})
            for key, data in type_new_metadata.items():
                endpoints = metadata_lookup(key)
                if endpoints:
                    endpoint = endpoints[0]
                    if metadata_type not in endpoint.metadata:
                        endpoint.metadata[metadata_type] = defaultdict(dict)
                    if key in endpoint.metadata[metadata_type]:
                        endpoint.metadata[metadata_type][key].update(data)
                    else:
                        endpoint.metadata[metadata_type][key] = data
                    updated.add(endpoint)
        return updated

    def format_rabbit_message(self, item, faucet_event, remove_list):
        '''
        read a message off the rabbit_q
        the message should be item = (routing_key,msg)
        '''
        routing_key, my_obj = item
        self.logger.debug(
            'routing_key: {0} rabbit_message: {1}'.format(routing_key, my_obj))

        def handler_algos_decider(my_obj):
            self.logger.debug('decider value:{0}'.format(my_obj))
            data = my_obj.get('data', None)
            tool = my_obj.get('tool', None)
            if isinstance(data, dict) and data:
                new_metadata = data
                # this is obsolete
                # if tool == 'p0f':
                #    ip_metadata = {}
                #    for ip, ip_data in data.items():
                #        if ip_data and ip_data.get('full_os', None):
                #            ip_metadata[ip] = ip_data
                #    new_metadata = {'ipv4_addresses': ip_metadata}
                # elif tool == 'networkml':
                #    mac_metadata = {}
                #    for name, message in data.items():
                #        if name == 'pcap':
                #            continue
                #        if message.get('valid', False):
                #            source_mac = message.get('source_mac', None)
                #            if source_mac:
                #                mac_metadata[source_mac] = message
                #    new_metadata = {'mac_addresses': mac_metadata}
                # Generic handler for future tools.
                updated = self.merge_metadata(new_metadata)
                if updated:
                    for endpoint in updated:
                        if endpoint.operation_active():
                            self.sdnc.unmirror_endpoint(endpoint)
                    return data
            return {}

        def handler_action_ignore(my_obj):
            for name in my_obj:
                endpoint = self.sdnc.endpoints.get(name, None)
                if endpoint:
                    endpoint.ignore = True
            return {}

        def handler_action_clear_ignored(my_obj):
            for name in my_obj:
                endpoint = self.sdnc.endpoints.get(name, None)
                if endpoint:
                    endpoint.ignore = False
            return {}

        def handler_action_change(my_obj):
            for name, state in my_obj:
                endpoint = self.sdnc.endpoints.get(name, None)
                if endpoint:
                    try:
                        if endpoint.operation_active():
                            self.sdnc.unmirror_endpoint(endpoint)
                        # pytype: disable=attribute-error
                        endpoint.machine_trigger(state)
                        endpoint.p_next_state = None
                        if endpoint.operation_active():
                            self.sdnc.mirror_endpoint(endpoint)
                            self.prom.prom_metrics['ncapture_count'].inc()
                    except Exception as e:  # pragma: no cover
                        self.logger.error(
                            'Unable to change endpoint {0} because: {1}'.format(endpoint.name, str(e)))
            return {}

        def handler_action_update_acls(my_obj):
            for ip in my_obj:
                rules = my_obj[ip]
                endpoints = self.sdnc.endpoints_by_ip(ip)
                if endpoints:
                    endpoint = endpoints[0]
                    try:
                        status = Actions(
                            endpoint, self.sdnc.sdnc).update_acls(
                                rules_file=self.config['RULES_FILE'], endpoints=endpoints, force_apply_rules=rules)
                        if not status:
                            self.logger.warning(
                                'Unable to apply rules: {0} to endpoint: {1}'.format(rules, endpoint.name))
                    except Exception as e:
                        self.logger.error(
                            'Unable to apply rules: {0} to endpoint: {1} because {2}'.format(rules, endpoint.name, str(e)))
            return {}

        def handler_action_remove(my_obj):
            remove_list.extend([name for name in my_obj])
            return {}

        def handler_action_remove_ignored(_my_obj):
            remove_list.extend([
                endpoint.name for endpoint in self.sdnc.endpoints.values()
                if endpoint.ignore])
            return {}

        def handler_faucet_event(my_obj):
            if self.sdnc and self.sdnc.sdnc:
                faucet_event.append(my_obj)
                return my_obj
            return {}

        handlers = {
            'poseidon.algos.decider': handler_algos_decider,
            'poseidon.action.ignore': handler_action_ignore,
            'poseidon.action.clear.ignored': handler_action_clear_ignored,
            'poseidon.action.change': handler_action_change,
            'poseidon.action.update_acls': handler_action_update_acls,
            'poseidon.action.remove': handler_action_remove,
            'poseidon.action.remove.ignored': handler_action_remove_ignored,
            self.config['FA_RABBIT_ROUTING_KEY']: handler_faucet_event,
        }

        handler = handlers.get(routing_key, None)
        if handler is not None:
            ret_val = handler(my_obj)
            return ret_val, True

        self.logger.error(
            'no handler for routing_key {0}'.format(routing_key))
        return {}, False

    def update_routing_key_time(self, routing_key):
        if self.prom:
            self.prom.prom_metrics['last_rabbitmq_routing_key_time'].labels(
                routing_key=routing_key).set(time.time())

    def handle_rabbit(self):
        events = 0
        faucet_event = []
        remove_list = []
        while True:
            found_work, rabbit_msg = self.prom.runtime_callable(
                partial(self.get_q_item, self.m_queue))
            if not found_work:
                break
            events += 1
            # faucet_event and remove_list get updated as references because partial()
            self.prom.runtime_callable(
                partial(self.format_rabbit_message, rabbit_msg, faucet_event, remove_list))
        return (events, faucet_event, remove_list)

    def ignore_rabbit(self, routing_key, body):
        ''' drop ignored messages. '''
        if routing_key == self.config['FA_RABBIT_ROUTING_KEY']:
            if self.sdnc and self.sdnc.sdnc:
                if self.sdnc.sdnc.ignore_event(body):
                    return True
        return False

    def rabbit_callback(self, ch, method, _properties, body, q=None):
        ''' callback, places rabbit data into internal queue'''
        body = json.loads(body)
        self.logger.debug('got a message: {0}:{1} (qsize {2})'.format(
            method.routing_key, body, q.qsize()))
        if q is not None:
            self.update_routing_key_time(method.routing_key)
            if not self.ignore_rabbit(method.routing_key, body):
                q.put((method.routing_key, body))
        ch.basic_ack(delivery_tag=method.delivery_tag)

    def process(self, monitor):
        while True:
            events, faucet_event, remove_list = self.prom.runtime_callable(
                self.handle_rabbit)
            if remove_list:
                for endpoint_name in remove_list:
                    if endpoint_name in self.sdnc.endpoints:
                        del self.sdnc.endpoints[endpoint_name]
            if faucet_event:
                self.prom.runtime_callable(
                    partial(self.sdnc.check_endpoints, faucet_event))
            # schedule_mirroring should be abstracted out
            events += self.prom.runtime_callable(monitor.schedule_mirroring)
            found_work, schedule_func = self.prom.runtime_callable(
                partial(self.get_q_item, self.job_queue))
            if found_work and callable(schedule_func):
                events += self.prom.runtime_callable(schedule_func)
            if events:
                self.prom.update_endpoint_metadata(self.sdnc.endpoints)
            time.sleep(1)

    @staticmethod
    def get_q_item(q):
        '''
        attempt to get a work item from the queue
        m_queue -> (routing_key, body)
        a read from get_q_item should be of the form
        (boolean,(routing_key, body))
        '''
        try:
            item = q.get_nowait()
            q.task_done()
            return (True, item)
        except queue.Empty:  # pragma: no cover
            pass

        return (False, None)
